#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library
import decimal
import os

import octobot_trading.enums as enums
import octobot_commons.enums as commons_enums
import octobot_commons.constants as commons_constants
import octobot_commons.os_util as os_util

# Strings
CURRENT_PORTFOLIO_STRING = "Current Portfolio :"
CONFIG_PORTFOLIO_FREE = "free"
CONFIG_PORTFOLIO_USED = "used"
CONFIG_PORTFOLIO_TOTAL = "total"
CONFIG_PORTFOLIO_MARGIN = "margin"
REAL_TRADER_STR = "[Real Trader] "
SIMULATOR_TRADER_STR = "[Simulator] "

# Trader
DEFAULT_REFERENCE_MARKET = "BTC"
CURRENCY_DEFAULT_MAX_PRICE_DIGITS = 8
ALLOW_SIMULATED_ORDERS_INSTANT_FILL = os_util.parse_boolean_environment_var(
    "ALLOW_SIMULATED_ORDERS_INSTANT_FILL", "False"
)

# Order creation
ORDER_DATA_FETCHING_TIMEOUT = 5 * commons_constants.MINUTE_TO_SECONDS
ACTIVE_ORDER_STRATEGY_SWAP_TIMEOUT = 2 * commons_constants.MINUTE_TO_SECONDS
CHAINED_ORDER_PRICE_FETCHING_TIMEOUT = 1    # should be instant or ignored
CHAINED_ORDERS_OUTDATED_PRICE_ALLOWANCE = decimal.Decimal("0.005")  # allows 0.5% outdated price error
# create instantly filled limit orders 0.5% beyond market
INSTANT_FILLED_LIMIT_ORDER_PRICE_DELTA = decimal.Decimal("0.005")
CREATED_ORDER_FORCED_UPDATE_PERIOD = 5
DEFAULT_MAX_DEFAULT_ORDERS_COUNT = 100
DEFAULT_MAX_STOP_ORDERS_COUNT = 10
INCLUDE_DUSTS_IN_SELL_ORDERS_WHEN_POSSIBLE = os_util.parse_boolean_environment_var(
    "INCLUDE_DUSTS_IN_SELL_ORDERS_WHEN_POSSIBLE", "true"
)

# Portfolio
MAX_PORTFOLIO_SYNC_ATTEMPTS = 1
SYNC_ATTEMPTS_INTERVAL = 20
SUB_PORTFOLIO_ALLOWED_MISSING_RATIO = decimal.Decimal("0.01")   # Allow 1% missing funds
SUB_PORTFOLIO_ALLOWED_DELTA_RATIO = decimal.Decimal("0.05")   # Allow 5% delta compared to filled orders
MAX_ORDER_INFERENCE_QUICK_CHECK_COMBINATIONS_COUNT = 10000
MAX_ORDER_SECONDARY_INFERENCE_COMBINATIONS_COUNT = 500000 # no more than 500.000 combinations to check to avoid overloads
MAX_NO_THREAD_WORSE_CASE_SCENARIO_FULLY_HANLDED_INFERENCE = 15 # 7 filled orders out of 15 is 6.435 combinations, 16 is 12.870 combinations (>10.000)
MAX_ORDERS_WORSE_CASE_SCENARIO_FULLY_HANLDED_INFERENCE = 21 # 10(or 11) filled orders out of 21 is 352.716 combinations, 22 is 705.432 combinations (>500.000)
MAX_ORDER_INFERENCE_ITERATIONS_DURATION = 1
ORDER_INFERENCE_SLEEP_TIME = 1

# Tentacles
TRADING_MODE_REQUIRED_STRATEGIES = "required_strategies"
TRADING_MODE_REQUIRED_STRATEGIES_MIN_COUNT = "required_strategies_min_count"
TENTACLES_TRADING_MODE_PATH = "Mode"
CONFIG_CANDLES_HISTORY_SIZE_TITLE = "Candles history size"
CONFIG_CANDLES_HISTORY_SIZE_KEY = CONFIG_CANDLES_HISTORY_SIZE_TITLE.replace(" ", "_")
CONFIG_BUY_ORDER_AMOUNT = "buy_order_amount"
CONFIG_SELL_ORDER_AMOUNT = "sell_order_amount"
CONFIG_LEVERAGE = "leverage"
TRADING_MODE_ACTIVITY_REASON = "reason"

# Exchange
DEFAULT_EXCHANGE_TIME_LAG = 10
DEFAULT_BACKTESTING_TIME_LAG = 0
INFINITE_MAX_HANDLED_PAIRS_WITH_TIMEFRAME = -1
DEFAULT_CANDLE_HISTORY_SIZE = 200
NO_DATA_LIMIT = -1
DEFAULT_FAILED_REQUEST_RETRY_TIME = 1
FAILED_NETWORK_REQUEST_RETRY_ATTEMPTS = 5
FAILED_PROXY_NETWORK_REQUEST_RETRY_ATTEMPTS = 2
TOOLS_FAILED_NETWORK_REQUEST_ATTEMPTS = int(os.getenv("TOOLS_FAILED_NETWORK_REQUEST_ATTEMPTS", "2"))
TOOLS_FAILED_NETWORK_REQUEST_RETRY_DELAY = int(os.getenv("TOOLS_FAILED_NETWORK_REQUEST_RETRY_DELAY", "5"))
DEFAULT_REQUEST_TIMEOUT = int(os.getenv("DEFAULT_REQUEST_TIMEOUT", "20000"))    # default ccxt is 10s, use 20
ENABLE_EXCHANGE_HTTP_PROXY_FROM_ENV = os_util.parse_boolean_environment_var(
    "ENABLE_EXCHANGE_HTTP_PROXY_FROM_ENV", "True"
)
EXCHANGE_HTTP_PROXY_AUTHENTICATED_URL = os.getenv("EXCHANGE_HTTP_PROXY_AUTHENTICATED_URL")
EXCHANGE_HTTPS_PROXY_AUTHENTICATED_URL = os.getenv("EXCHANGE_HTTPS_PROXY_AUTHENTICATED_URL")
EXCHANGE_SOCKS_PROXY_AUTHENTICATED_URL = os.getenv("EXCHANGE_SOCKS_PROXY_AUTHENTICATED_URL")
ENABLE_CCXT_VERBOSE = os_util.parse_boolean_environment_var("ENABLE_CCXT_VERBOSE", "False")
ENABLE_CCXT_RATE_LIMIT = os_util.parse_boolean_environment_var("ENABLE_CCXT_RATE_LIMIT", "True")
ENABLE_CCXT_REQUESTS_COUNTER = os_util.parse_boolean_environment_var("ENABLE_CCXT_REQUESTS_COUNTER", "False")
FETCH_MIN_EXCHANGE_MARKETS = os_util.parse_boolean_environment_var("FETCH_MIN_EXCHANGE_MARKETS", "False")
CCXT_DEFAULT_CACHE_LIMIT = int(os.getenv("CCXT_DEFAULT_CACHE_LIMIT", "1000"))  # 1000: default ccxt value
CCXT_TRADES_CACHE_LIMIT = int(os.getenv("CCXT_TRADES_CACHE_LIMIT", str(CCXT_DEFAULT_CACHE_LIMIT)))
CCXT_ORDERS_CACHE_LIMIT = int(os.getenv("CCXT_ORDERS_CACHE_LIMIT", str(CCXT_DEFAULT_CACHE_LIMIT)))
CCXT_OHLCV_CACHE_LIMIT = int(os.getenv("CCXT_OHLCV_CACHE_LIMIT", str(CCXT_DEFAULT_CACHE_LIMIT)))
CCXT_WATCH_ORDER_BOOK_LIMIT = int(os.getenv("CCXT_WATCH_ORDER_BOOK_LIMIT", str(CCXT_DEFAULT_CACHE_LIMIT)))
CCXT_TIMEOUT_ON_EXIT_MS = 100
THROTTLED_WS_UPDATES = float(os.getenv("THROTTLED_WS_UPDATES", "0.1"))  # avoid spamming CPU
MAX_CANDLES_IN_RAM = int(os.getenv("MAX_CANDLES_IN_RAM", "3000"))    # max candles per CandlesManager
STORAGE_ORIGIN_VALUE = "origin_value"
DISPLAY_TIME_FRAME = commons_enums.TimeFrames.ONE_HOUR
DEFAULT_SUBACCOUNT_ID = "default_subaccount_id"
DEFAULT_ACCOUNT_ID = "default_account_id"

# Storage
ENABLE_LIVE_CANDLES_STORAGE = os_util.parse_boolean_environment_var("ENABLE_LIVE_CANDLES_STORAGE", "False")
ENABLE_HISTORICAL_ORDERS_UPDATES_STORAGE = os_util.parse_boolean_environment_var("ENABLE_HISTORICAL_ORDERS_UPDATES_STORAGE", "False")
ENABLE_SIMULATED_ORDERS_STORAGE = os_util.parse_boolean_environment_var("ENABLE_SIMULATED_ORDERS_STORAGE", "False")
AUTH_UPDATE_DEBOUNCE_DURATION = float(os.getenv("AUTH_UPDATE_DEBOUNCE_DURATION", "10"))

# Decimal default values (decimals are immutable, can be stored as constant)
ZERO = decimal.Decimal(0)
ONE = decimal.Decimal(1)
ONE_HUNDRED = decimal.Decimal(100)
NaN = decimal.Decimal("nan")

# exchanges where test_get_historical_symbol_prices is successful can be listed here
FULL_CANDLE_HISTORY_EXCHANGES = [
    "ascendex",
    "binance",
    "bitfinex",
    "bitstamp",
    "bybit",
    "bingx",
    "hollaex",
    "htx",
    "kucoin",
    "okcoin",
    "okx",
    "myokx",
    "okxus",
    "mexc",
    "coinbase",
    "binanceus",
    "bitmart",
    "bitmex",
]

DEFAULT_FUTURE_EXCHANGES = ["binanceusdm", "bybit"]
TESTED_EXCHANGES = [
    "binance",
    "kucoin",
    "okx",
] + sorted([
    "binanceus",
    "coinbase",
    "cryptocom",
    "htx",
    "hyperliquid",
    "bitget",
    "gateio",
    "ascendex",
    "bybit",
    "phemex",
    "hollaex",
    "mexc",
    "bingx",
    "coinex",
    "bitmart",
])
DEFAULT_FUTURE_EXCHANGES = sorted(["bybit"])
SIMULATOR_TESTED_EXCHANGES = sorted(["bitfinex", "bithumb", "bitstamp", "bitmex",
                              "hitbtc", "kraken", "poloniex", "bitso", "ndax", "upbit",
                              "myokx", "okxus",
                              "wavesexchange",])


# exchanges
CONFIG_DEFAULT_FEES = 0.001
CONFIG_DEFAULT_SIMULATOR_FEES = 0
FEES_SAFETY_MARGIN = decimal.Decimal("1.25")    # allow 25% error margin when simulating fees

DEFAULT_SYMBOL_LEVERAGE = ONE
DEFAULT_SYMBOL_MAX_LEVERAGE = ONE_HUNDRED
DEFAULT_SYMBOL_MARGIN_TYPE = enums.MarginType.ISOLATED
DEFAULT_SYMBOL_CONTRACT_TYPE = enums.FutureContractType.LINEAR_PERPETUAL
DEFAULT_SYMBOL_CONTRACT_SIZE = ONE
DEFAULT_SYMBOL_POSITION_MODE = enums.PositionMode.ONE_WAY
DEFAULT_SYMBOL_FUNDING_RATE = decimal.Decimal("0.00005")
DEFAULT_SYMBOL_MAINTENANCE_MARGIN_RATE = decimal.Decimal("0.01")
RETRIABLE_EXCHANGE_ERRORS_DESC: set[str] = set(os.getenv(
    "RETRIABLE_EXCHANGE_ERRORS_DESC", (
        'Internal Server Error:Bad gateway:socket hang up'
        ':read ECONNRESET:read ETIMEDOUT'
    )
).split(":"))

# exchange proxy
RETRIABLE_EXCHANGE_PROXY_ERRORS_DESC: set[str] = set(os.getenv(
    "RETRIABLE_EXCHANGE_PROXY_ERRORS_DESC", "message='Service Unavailable'"
).split(":"))

# used to force margin type update before positions init (if necessary)
FORCED_MARGIN_TYPE = enums.MarginType(os.getenv("FORCED_MARGIN_TYPE", enums.MarginType.ISOLATED.value))

# API
API_LOGGER_TAG = "TradingApi"

# Channels
# Exchange public data
TICKER_CHANNEL = "Ticker"
MINI_TICKER_CHANNEL = "MiniTicker"
RECENT_TRADES_CHANNEL = "RecentTrade"
LIQUIDATIONS_CHANNEL = "Liquidations"
ORDER_BOOK_CHANNEL = "OrderBook"
ORDER_BOOK_TICKER_CHANNEL = "OrderBookTicker"
KLINE_CHANNEL = "Kline"
OHLCV_CHANNEL = "OHLCV"
MARK_PRICE_CHANNEL = "MarkPrice"
FUNDING_CHANNEL = "Funding"

# Exchange personal data
TRADES_CHANNEL = "Trades"
ORDERS_CHANNEL = "Orders"
BALANCE_CHANNEL = "Balance"
BALANCE_PROFITABILITY_CHANNEL = "BalanceProfitability"
POSITIONS_CHANNEL = "Positions"
INDIVIDUAL_ORDER_SYNC_TIMEOUT = 1 * commons_constants.MINUTE_TO_SECONDS
MAX_TRADES_COUNT = int(os.getenv("MAX_TRADES_COUNT", "10000"))    # larger values can use a large part of ram

# History
DEFAULT_SAVED_HISTORICAL_TIMEFRAMES = [commons_enums.TimeFrames.ONE_DAY]
HISTORICAL_CANDLES_FETCH_DEFAULT_TIMEOUT = 30

# 946742400 is 01/01/2000, if trade time is lower, there is an issue.
MINIMUM_VAL_TRADE_TIME = 946688400

# Internal
MODE_CHANNEL = "Mode"

WEBSOCKET_FEEDS_TO_TRADING_CHANNELS = {
    TICKER_CHANNEL: [enums.WebsocketFeeds.TICKER],
    MINI_TICKER_CHANNEL: [enums.WebsocketFeeds.MINI_TICKER],
    RECENT_TRADES_CHANNEL: [enums.WebsocketFeeds.TRADES],
    LIQUIDATIONS_CHANNEL: [enums.WebsocketFeeds.LIQUIDATIONS],
    ORDER_BOOK_CHANNEL: [enums.WebsocketFeeds.L2_BOOK, enums.WebsocketFeeds.L3_BOOK],
    ORDER_BOOK_TICKER_CHANNEL: [enums.WebsocketFeeds.BOOK_TICKER],
    KLINE_CHANNEL: [enums.WebsocketFeeds.KLINE],
    OHLCV_CHANNEL: [enums.WebsocketFeeds.CANDLE],
    TRADES_CHANNEL: [enums.WebsocketFeeds.TRADE],
    ORDERS_CHANNEL: [enums.WebsocketFeeds.ORDERS],
    MARK_PRICE_CHANNEL: [enums.WebsocketFeeds.MARK_PRICE],
    BALANCE_CHANNEL: [enums.WebsocketFeeds.PORTFOLIO],
    POSITIONS_CHANNEL: [enums.WebsocketFeeds.POSITION],
    FUNDING_CHANNEL: [enums.WebsocketFeeds.FUNDING]
}

ALWAYS_STARTED_REST_PRODUCER_CHANNELS = [
    TICKER_CHANNEL, # use to force mark price update when necessary: should always be reachable
]

FILL_ORDER_STATUS_SCOPE = [enums.OrderStatus.CLOSED,
                           enums.OrderStatus.FILLED,
                           enums.OrderStatus.PARTIALLY_FILLED]
CANCEL_ORDER_STATUS_SCOPE = [enums.OrderStatus.PENDING_CANCEL,
                             enums.OrderStatus.CANCELED,
                             enums.OrderStatus.EXPIRED,
                             enums.OrderStatus.REJECTED]

DEFAULT_INITIALIZATION_EVENT_TOPICS = [
    commons_enums.InitializationEventExchangeTopics.BALANCE,
    commons_enums.InitializationEventExchangeTopics.PROFITABILITY,
    commons_enums.InitializationEventExchangeTopics.ORDERS,
    commons_enums.InitializationEventExchangeTopics.TRADES,
    commons_enums.InitializationEventExchangeTopics.CANDLES,
    commons_enums.InitializationEventExchangeTopics.PRICE,
]

DEFAULT_FUTURES_INITIALIZATION_EVENT_TOPICS = DEFAULT_INITIALIZATION_EVENT_TOPICS + [
    commons_enums.InitializationEventExchangeTopics.POSITIONS,
    commons_enums.InitializationEventExchangeTopics.CONTRACTS,
    commons_enums.InitializationEventExchangeTopics.FUNDING,
]



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import copy
import typing

import octobot_trading.constants as constants
import octobot_trading.enums as enums
import octobot_trading.personal_data.orders.order as order_import
import octobot_trading.exchanges
import octobot_commons.symbols as commons_symbols


class Trade:
    CLOSING_TRADE_ORDER_STATUS = {enums.OrderStatus.CANCELED, enums.OrderStatus.FILLED, enums.OrderStatus.CLOSED}

    def __init__(self, trader):
        self.trader: octobot_trading.exchanges.Trader = trader
        self.exchange_manager: octobot_trading.exchanges.ExchangeManager = trader.exchange_manager

        self.status: enums.OrderStatus = enums.OrderStatus.OPEN
        self.creation_time: float = self.exchange_manager.exchange.get_exchange_current_time()

        self.trade_id: str = trader.parse_order_id(None)
        self.origin_order_id: str = None # type: ignore
        self.exchange_order_id: str = None # type: ignore
        # One order might create multiple trades when matched to multiple open orders.
        # in this case those trades would share the same exchange_order_id
        self.exchange_trade_id: str = None # type: ignore
        self.simulated: bool = True
        self.is_closing_order: bool = False
        self.is_from_this_octobot: bool = True

        self.symbol: str = None # type: ignore
        self.currency: typing.Optional[str] = None
        self.market: typing.Optional[str] = None
        self.taker_or_maker: str = None # type: ignore
        self.origin_price: decimal.Decimal = constants.ZERO
        self.origin_quantity: decimal.Decimal = constants.ZERO
        self.trade_type: typing.Optional[enums.TradeOrderType] = None
        self.side: enums.TradeOrderSide = None # type: ignore
        self.executed_quantity: decimal.Decimal = constants.ZERO
        self.canceled_time: float = 0
        self.executed_time: float = 0
        self.fee: typing.Optional[dict[str, typing.Any]] = None
        self.executed_price: decimal.Decimal = constants.ZERO
        self.trade_profitability: decimal.Decimal = constants.ZERO
        self.total_cost: decimal.Decimal = constants.ZERO
        self.reduce_only: bool = False
        self.tag: str = None # type: ignore
        self.quantity_currency: str = None # type: ignore
        self.associated_entry_ids: typing.Optional[list[str]] = None
        self.broker_applied: bool = False

        # raw exchange trade type, used to create trade dict
        self.exchange_trade_type: typing.Optional[enums.TradeOrderType] = None

    def update_from_order(self, order, creation_time=0, canceled_time=0, executed_time=0, exchange_trade_id=None):
        self.currency = order.currency
        self.market = order.market
        self.taker_or_maker = order.taker_or_maker
        self.executed_quantity = order.filled_quantity
        self.executed_price = order.filled_price
        self.origin_price = order.origin_price
        self.origin_quantity = order.origin_quantity
        self.total_cost = order.total_cost
        self.quantity_currency = order.quantity_currency
        self.trade_type = order.order_type
        self.exchange_trade_type = order.exchange_order_type
        self.status = order.status
        self.fee = order.fee
        self.trade_id = order.order_id
        self.origin_order_id = order.order_id
        self.exchange_order_id = order.exchange_order_id
        self.exchange_trade_id = exchange_trade_id or self.exchange_trade_id
        self.simulated = order.simulated
        self.side = order.side
        self.creation_time = order.creation_time if order.creation_time > 0 else creation_time
        self.canceled_time = order.canceled_time if order.canceled_time > 0 else canceled_time
        self.executed_time = order.executed_time if order.executed_time > 0 else executed_time
        self.symbol = order.symbol
        self.is_closing_order = order.status in self.CLOSING_TRADE_ORDER_STATUS
        self.is_from_this_octobot = order.is_from_this_octobot
        self.reduce_only = order.reduce_only
        self.tag = order.tag
        self.associated_entry_ids = order.associated_entry_ids
        self.broker_applied = order.broker_applied

    def get_time(self):
        return self.executed_time if self.has_been_executed() else self.canceled_time

    def get_quantity(self):
        return self.executed_quantity if self.has_been_executed() else self.origin_quantity

    def has_been_executed(self):
        return self.status is not enums.OrderStatus.CANCELED and self.status is not enums.OrderStatus.EXPIRED

    def to_dict(self):
        return {
            enums.ExchangeConstantsOrderColumns.ID.value: self.trade_id,
            enums.ExchangeConstantsOrderColumns.ORDER_ID.value: self.origin_order_id,
            enums.ExchangeConstantsOrderColumns.EXCHANGE_ID.value: self.exchange_order_id,
            enums.ExchangeConstantsOrderColumns.EXCHANGE_TRADE_ID.value: self.exchange_trade_id,
            enums.ExchangeConstantsOrderColumns.SYMBOL.value: self.symbol,
            enums.ExchangeConstantsOrderColumns.MARKET.value: self.market,
            enums.ExchangeConstantsOrderColumns.PRICE.value: self.executed_price,
            enums.ExchangeConstantsOrderColumns.STATUS.value: self.status.value,
            enums.ExchangeConstantsOrderColumns.TIMESTAMP.value: self.get_time(),
            enums.ExchangeConstantsOrderColumns.TYPE.value: self.exchange_trade_type.value,
            enums.ExchangeConstantsOrderColumns.SIDE.value: self.side.value,
            enums.ExchangeConstantsOrderColumns.AMOUNT.value: self.get_quantity(),
            enums.ExchangeConstantsOrderColumns.COST.value: self.total_cost,
            enums.ExchangeConstantsOrderColumns.QUANTITY_CURRENCY.value: self.quantity_currency,
            enums.ExchangeConstantsOrderColumns.TAKER_OR_MAKER.value: self.taker_or_maker,
            enums.ExchangeConstantsOrderColumns.FEE.value: self.fee,
            enums.ExchangeConstantsOrderColumns.REDUCE_ONLY.value: self.reduce_only,
            enums.ExchangeConstantsOrderColumns.TAG.value: self.tag,
            enums.ExchangeConstantsOrderColumns.ENTRIES.value: self.associated_entry_ids,
            enums.ExchangeConstantsOrderColumns.BROKER_APPLIED.value: self.broker_applied,
            enums.TradeExtraConstants.CREATION_TIME.value: self.creation_time,
        }

    @classmethod
    def from_dict(cls, trader, trade_dict):
        trade = cls(trader)
        trade.trade_id = trade_dict.get(enums.ExchangeConstantsOrderColumns.ID.value, trade.trade_id)
        trade.origin_order_id = trade_dict.get(enums.ExchangeConstantsOrderColumns.ORDER_ID.value)
        trade.exchange_trade_id = trade_dict.get(enums.ExchangeConstantsOrderColumns.EXCHANGE_TRADE_ID.value)
        trade.exchange_order_id = trade_dict.get(enums.ExchangeConstantsOrderColumns.EXCHANGE_ID.value)
        trade.symbol = trade_dict.get(enums.ExchangeConstantsOrderColumns.SYMBOL.value)
        trade.currency, trade.market = commons_symbols.parse_symbol(trade.symbol).base_and_quote()
        trade.market = trade_dict.get(enums.ExchangeConstantsOrderColumns.MARKET.value)
        trade.executed_price = decimal.Decimal(str(trade_dict.get(enums.ExchangeConstantsOrderColumns.PRICE.value)))
        trade.status = enums.OrderStatus(trade_dict.get(enums.ExchangeConstantsOrderColumns.STATUS.value,
                                                        enums.OrderStatus.CLOSED.value))
        if trade.has_been_executed():
            trade.executed_time = trade_dict.get(enums.ExchangeConstantsOrderColumns.TIMESTAMP.value)
            trade.executed_quantity = \
                decimal.Decimal(str(trade_dict.get(enums.ExchangeConstantsOrderColumns.AMOUNT.value)))
        else:
            trade.canceled_time = trade_dict.get(enums.ExchangeConstantsOrderColumns.TIMESTAMP.value)
            trade.origin_quantity = \
                decimal.Decimal(str(trade_dict.get(enums.ExchangeConstantsOrderColumns.AMOUNT.value)))
        trade.exchange_trade_type = enums.TradeOrderType(trade_dict.get(enums.ExchangeConstantsOrderColumns.TYPE.value))
        trade.side = enums.TradeOrderSide(trade_dict.get(enums.ExchangeConstantsOrderColumns.SIDE.value))
        trade.trade_type = order_import.parse_order_type({
            enums.ExchangeConstantsOrderColumns.SIDE.value: trade.side.value,
            enums.ExchangeConstantsOrderColumns.TYPE.value: trade.exchange_trade_type.value,
        })[1]
        trade.total_cost = decimal.Decimal(str(trade_dict.get(enums.ExchangeConstantsOrderColumns.COST.value)))
        trade.quantity_currency = trade_dict.get(enums.ExchangeConstantsOrderColumns.QUANTITY_CURRENCY.value)
        trade.taker_or_maker = trade_dict.get(enums.ExchangeConstantsOrderColumns.TAKER_OR_MAKER.value)
        trade.fee = copy.copy(trade_dict.get(enums.ExchangeConstantsOrderColumns.FEE.value))
        if trade.fee and enums.FeePropertyColumns.COST.value in trade.fee:
            trade.fee[enums.FeePropertyColumns.COST.value] = \
                decimal.Decimal(str(trade.fee[enums.FeePropertyColumns.COST.value]))
        trade.reduce_only = trade_dict.get(enums.ExchangeConstantsOrderColumns.REDUCE_ONLY.value)
        trade.tag = trade_dict.get(enums.ExchangeConstantsOrderColumns.TAG.value)
        trade.associated_entry_ids = trade_dict.get(enums.ExchangeConstantsOrderColumns.ENTRIES.value)
        trade.broker_applied = trade_dict.get(
            enums.ExchangeConstantsOrderColumns.BROKER_APPLIED.value, trade.broker_applied
        )
        trade.creation_time = trade_dict.get(enums.TradeExtraConstants.CREATION_TIME.value)
        return trade

    def clear(self):
        self.trader = None # type: ignore
        self.exchange_manager = None # type: ignore



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library

import asyncio
import contextlib

import octobot_commons.channels_name as channels_name
import octobot_trading.exchanges as exchanges
import octobot_trading.util.test_tools.exchanges_test_tools as exchanges_test_tools

try:
    import mock
except ImportError:
    pass

EXPECTED_PUSHED_CHANNELS = {
    channels_name.OctoBotTradingChannelsName.ORDER_BOOK_TICKER_CHANNEL.value,
    channels_name.OctoBotTradingChannelsName.ORDER_BOOK_CHANNEL.value,
    channels_name.OctoBotTradingChannelsName.RECENT_TRADES_CHANNEL.value,
    channels_name.OctoBotTradingChannelsName.MINI_TICKER_CHANNEL.value,
    channels_name.OctoBotTradingChannelsName.OHLCV_CHANNEL.value,
    channels_name.OctoBotTradingChannelsName.KLINE_CHANNEL.value,
    channels_name.OctoBotTradingChannelsName.TICKER_CHANNEL.value
}


@contextlib.asynccontextmanager
async def ws_exchange_manager(config: object, exchange_name: str):
    exchange_manager_instance = None
    try:
        exchange_manager_instance = await exchanges_test_tools.create_test_exchange_manager(
            config=config, exchange_name=exchange_name)
        yield exchange_manager_instance
    finally:
        await exchanges_test_tools.stop_test_exchange_manager(exchange_manager_instance)


async def test_unauthenticated_push_to_channel_coverage_websocket(
        websocket_exchange_class: exchanges.AbstractWebsocketExchange,
        websocket_connector_class,
        exchange_manager: exchanges.ExchangeManager,
        config: object,
        symbols: list,
        time_frames: list,
        expected_pushed_channels: set = None,
        time_before_assert: int = 120):
    """
    DEPRECATED: TODO udpate for ccxt ws
    """

    # To fix `TypeError: can't set attributes of built-in/extension type 'datetime.date'`
    class TestAbstractWebsocketExchange(websocket_exchange_class):
        pass

    pushed_channel_names = set()
    if expected_pushed_channels is None:
        expected_pushed_channels = EXPECTED_PUSHED_CHANNELS

    async def mock_push_to_channel(self, channel_name, **kwargs):
        pushed_channel_names.add(channel_name)

    with mock.patch.object(websocket_connector_class,
                           'push_to_channel',
                           new=mock_push_to_channel):
        ws_exchange = TestAbstractWebsocketExchange(config, exchange_manager)
        with mock.patch.object(TestAbstractWebsocketExchange,
                               'get_exchange_connector_class',
                               new=mock.Mock()) as get_exchange_connector_class_mock:
            get_exchange_connector_class_mock.return_value = websocket_connector_class
            await ws_exchange.init_websocket(time_frames=time_frames,
                                             trader_pairs=symbols,
                                             tentacles_setup_config=None)
            await ws_exchange.start_sockets()
            await asyncio.sleep(time_before_assert)

    try:
        assert pushed_channel_names == expected_pushed_channels
    finally:
        await ws_exchange.close_sockets()
        ws_exchange.clear()



# pylint: disable=E0611
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio

import octobot_commons.async_job as async_job
import octobot_commons.tree as commons_tree
import octobot_commons.enums as commons_enums
import octobot_commons.html_util as html_util

import octobot_trading.errors as errors
import octobot_trading.personal_data.orders.channel.orders as orders_channel
import octobot_trading.constants as constants


class OrdersUpdater(orders_channel.OrdersProducer):
    """
    Update open and close orders from exchange
    Can also be used to update a specific order from exchange
    """

    CHANNEL_NAME = constants.ORDERS_CHANNEL
    ORDERS_UPDATE_LIMIT = None
    ORDERS_STARTING_REFRESH_TIME = 10
    OPEN_ORDER_REFRESH_TIME = 7
    CLOSE_ORDER_REFRESH_TIME = 81
    TIME_BETWEEN_ORDERS_REFRESH = 2
    DEPENDENCIES_TIMEOUT = 30
    OPEN_ORDER_INITIAL_FETCH_GIVE_UP_TIMEOUT = 30

    def __init__(self, channel):
        super().__init__(channel)

        self._is_initialized_event_set = False
        # create async jobs
        self.open_orders_job = async_job.AsyncJob(self._open_orders_fetch_and_push,
                                                  execution_interval_delay=self.OPEN_ORDER_REFRESH_TIME,
                                                  min_execution_delay=self.TIME_BETWEEN_ORDERS_REFRESH)
        self.closed_orders_job = async_job.AsyncJob(self._closed_orders_fetch_and_push,
                                                    execution_interval_delay=self.CLOSE_ORDER_REFRESH_TIME,
                                                    min_execution_delay=self.TIME_BETWEEN_ORDERS_REFRESH)
        self.order_update_job = async_job.AsyncJob(self._order_fetch_and_push,
                                                   is_periodic=False,
                                                   enable_multiple_runs=True)
        self.order_update_job.add_job_dependency(self.open_orders_job)
        self.open_orders_job.add_job_dependency(self.order_update_job)

    async def initialize(self) -> None:
        """
        Initialize data before starting jobs
        """
        try:
            await self.wait_for_dependencies(
                [
                    commons_tree.get_exchange_path(
                        self.channel.exchange_manager.exchange_name,
                        commons_enums.InitializationEventExchangeTopics.CONTRACTS.value
                    ),
                    commons_tree.get_exchange_path(
                        self.channel.exchange_manager.exchange_name,
                        commons_enums.InitializationEventExchangeTopics.POSITIONS.value
                    ),
                ],
                self.DEPENDENCIES_TIMEOUT
            )
            await self.fetch_and_push(is_from_bot=False, retry_till_success=True)
            await self._restore_required_virtual_orders()
        except errors.NotSupported:
            self.logger.error(f"{self.channel.exchange_manager.exchange_name} is not supporting open orders updates")
            await self.pause()
        except Exception as e:
            self.logger.exception(e, True, f"Fail to initialize orders : {html_util.get_html_summary_if_relevant(e)}")

    async def start(self) -> None:
        """
        Start updater jobs
        """
        await self.initialize()
        await asyncio.sleep(self.ORDERS_STARTING_REFRESH_TIME)
        await self.open_orders_job.run(retry_attempts=1)
        # await self.closed_orders_job.run()

    async def fetch_and_push(self, is_from_bot=True, limit=ORDERS_UPDATE_LIMIT, retry_till_success=False):
        """
        Update open and closed orders from exchange
        :param is_from_bot: True if the order was created by OctoBot
        :param limit: the exchange request orders count limit
        :param retry_till_success: retry request till it works. Should be rarely used as it might take some time
        """
        # should not raise: open orders are necessary
        try:
            await self._open_orders_fetch_and_push(is_from_bot=is_from_bot, limit=limit,
                                                   retry_till_success=retry_till_success)
        finally:
            if self.channel is not None:
                self.channel.exchange_manager.exchange_personal_data.on_completed_orders_fetch()
        await asyncio.sleep(self.TIME_BETWEEN_ORDERS_REFRESH)
        try:
            # can raise, closed orders are not critical data
            await self._closed_orders_fetch_and_push(limit=limit)
        except errors.NotSupported:
            self.logger.debug(f"{self.channel.exchange_manager.exchange_name} is not supporting closed orders updates")

    async def _open_orders_fetch_and_push(
        self, is_from_bot=True, limit=ORDERS_UPDATE_LIMIT, retry_till_success=False, retry_attempts=0
    ):
        """
        Update open orders from exchange
        :param is_from_bot: True if the order was created by OctoBot
        :param limit: the exchange request orders count limit
        :param retry_till_success: retry request till it works. Should be rarely used as it might take some time
        :param retry_attempts: how many times to retry before failing
        """
        for symbol in self.channel.exchange_manager.exchange_config.traded_symbol_pairs:
            if retry_till_success:
                open_orders: list = await self.channel.exchange_manager.exchange.retry_till_success(
                    self.OPEN_ORDER_INITIAL_FETCH_GIVE_UP_TIMEOUT,
                    self.channel.exchange_manager.exchange.get_open_orders, symbol=symbol, limit=limit,
                )
            elif retry_attempts:
                open_orders: list = await self.channel.exchange_manager.exchange.retry_n_time(
                    retry_attempts,
                    self.channel.exchange_manager.exchange.get_open_orders, symbol=symbol, limit=limit,
                )
            else:
                open_orders: list = await self.channel.exchange_manager.exchange.get_open_orders(
                    symbol=symbol, limit=limit
                )
            if open_orders:
                await self.push(open_orders, is_from_bot=is_from_bot)
            else:
                await self.handle_post_open_orders_update((symbol, ), open_orders, [], False, True)
            if not self._is_initialized_event_set:
                self._set_initialized_event(symbol)
        self._is_initialized_event_set = True

    def _set_initialized_event(self, symbol):
        # set init in updater as it's the only place we know if we fetched orders or not regardless of orders existence
        commons_tree.EventProvider.instance().trigger_event(
            self.channel.exchange_manager.bot_id, commons_tree.get_exchange_path(
                self.channel.exchange_manager.exchange_name,
                commons_enums.InitializationEventExchangeTopics.ORDERS.value,
                symbol=symbol
            )
        )

    async def _closed_orders_fetch_and_push(self, limit=ORDERS_UPDATE_LIMIT) -> None:
        """
        Update closed orders from exchange
        :param limit: the exchange request orders count limit
        """
        for symbol in self.channel.exchange_manager.exchange_config.traded_symbol_pairs:
            close_orders: list = await self.channel.exchange_manager.exchange.get_closed_orders(
                symbol=symbol, limit=limit)

            if close_orders:
                await self.push(close_orders, are_closed=True)

    async def update_order_from_exchange(self, order,
                                         should_notify=False,
                                         wait_for_refresh=False,
                                         force_job_execution=False,
                                         create_order_producer_if_missing=True):
        """
        Trigger order job refresh from exchange
        :param order: the order to update
        :param wait_for_refresh: if True, wait until the order refresh task to finish
        :param should_notify: if Orders channel consumers should be notified
        :param force_job_execution: When True, order_update_job will bypass its dependencies check
        :param create_order_producer_if_missing: Should be set to False when called by self to prevent spamming
        :return: True if the order was updated
        """
        await self.order_update_job.run(force=True, wait_for_task_execution=wait_for_refresh,
                                        ignore_dependencies_check=force_job_execution,
                                        order=order, should_notify=should_notify)

    async def _order_fetch_and_push(self, order, should_notify=False):
        """
        Update Order from exchange
        :param order: the order to update
        :param should_notify: if Orders channel consumers should be notified
        :return: True if the order was updated
        """
        exchange_name = self.channel.exchange_manager.exchange_name
        self.logger.info(f"Requested update for {order} on {exchange_name}")
        raw_order = await self.channel.exchange_manager.exchange.get_order(order.exchange_order_id, order.symbol)

        if raw_order is not None:
            self.logger.info(f"Received update for {order} on {exchange_name}: {raw_order}")
            await self.channel.exchange_manager.exchange_personal_data.handle_order_update_from_raw(
                order.exchange_order_id, raw_order, should_notify=should_notify
            )
            self.logger.info(f"Completed update for {order} on {exchange_name}")
        else:
            self.logger.info(f"Can't received update for {order} on {exchange_name}: received order is None")

    async def stop(self) -> None:
        """
        Stop producer by stopping its jobs
        """
        await super().stop()
        self.open_orders_job.stop()
        self.closed_orders_job.stop()
        self.order_update_job.stop()

    async def resume(self) -> None:
        """
        Resume producer by restarting its jobs
        """
        await super().resume()
        if not self.is_running:
            await self.run()



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import json
import time

import pytest

from tests import event_loop
from octobot_commons.tests.test_config import load_test_config
from octobot_trading.enums import TraderOrderType, OrderStatus, FeePropertyColumns
from octobot_trading.constants import ZERO
from octobot_trading.exchanges.exchange_manager import ExchangeManager
from octobot_trading.personal_data.orders.order_factory import create_order_instance_from_raw
from octobot_trading.exchanges.traders.trader_simulator import TraderSimulator

from octobot_trading.personal_data.trades import create_trade_instance_from_raw, create_trade_from_order

# All test coroutines will be treated as marked.
from octobot_trading.api.exchange import cancel_ccxt_throttle_task

pytestmark = pytest.mark.asyncio


class TestTradeFactory:
    DEFAULT_SYMBOL = "BTC/USDT"
    EXCHANGE_MANAGER_CLASS_STRING = "binanceus"

    @staticmethod
    async def init_default():
        config = load_test_config()
        exchange_manager = ExchangeManager(config, TestTradeFactory.EXCHANGE_MANAGER_CLASS_STRING)
        await exchange_manager.initialize(exchange_config_by_exchange=None)

        trader = TraderSimulator(config, exchange_manager)
        await trader.initialize()

        return config, exchange_manager, trader

    @staticmethod
    async def stop(exchange_manager):
        cancel_ccxt_throttle_task()
        await exchange_manager.stop()

    async def test_create_trade_instance_from_raw(self):
        _, exchange_manager, trader = await self.init_default()

        with pytest.raises(KeyError):
            create_trade_instance_from_raw(trader, {})

        raw_trade = json.loads(
            """
            {	
              "info": {},
              "id": "12345-67890:09876/54321",
              "exchange_id": "plop",
              "exchange_trade_id": "plop-trade",
              "timestamp": 1502962946216,
              "datetime": "2017-08-17 12:42:48.000",
              "symbol": "ETH/BTC",
              "order": "12345-67890:09876/54321",
              "type": "limit",
              "side": "buy",
              "takerOrMaker": "taker",
              "price": 0.06917684,
              "amount": 1.5,
              "cost": 0.10376526,
              "fee": {
                "cost": 0.0015,
                "currency": "ETH",
                "rate": 0.002
              }
            }
            """)

        trade = create_trade_instance_from_raw(trader, raw_trade)

        assert trade.trade_id == '12345-67890:09876/54321'
        assert trade.origin_order_id == '12345-67890:09876/54321'
        assert trade.exchange_order_id == 'plop'
        assert trade.exchange_trade_id == 'plop-trade'
        assert trade.trade_type == TraderOrderType.BUY_LIMIT
        assert trade.symbol == 'ETH/BTC'
        assert trade.total_cost == decimal.Decimal(str(0.10376526))
        assert trade.executed_quantity == decimal.Decimal(str(1.5))
        assert trade.origin_price == decimal.Decimal(str(0.06917684))
        assert trade.executed_price == decimal.Decimal(str(0.06917684))
        assert trade.executed_time == 1502962946216
        assert trade.status == OrderStatus.FILLED
        assert trade.fee == {
            FeePropertyColumns.COST.value: decimal.Decimal("0.0015"),
            FeePropertyColumns.CURRENCY.value: "ETH",
            FeePropertyColumns.RATE.value: 0.002,
            FeePropertyColumns.IS_FROM_EXCHANGE.value: True
        }
        assert trade.is_closing_order is True
        assert trade.tag is None

        await self.stop(exchange_manager)

    async def test_create_trade_from_order(self):
        _, exchange_manager, trader = await self.init_default()

        # limit order
        raw_order = json.loads(
            """
            {
                "id":                "12345-67890:09876/54321",
                "exchange_id":       "12345-67890:09876/1111",
                "datetime":          "2017-08-17 12:42:48.000",
                "timestamp":          1502962946216,
                "lastTradeTimestamp": 1502962956216,
                "status":     "open",
                "symbol":     "BTC/USDT",
                "type":       "limit",
                "side":       "sell",
                "price":       7684,
                "amount":      1.5,
                "filled":      1.1,
                "remaining":   0.4,
                "cost":        0.076094524,
                "trades":    [],
                "fee": {
                    "currency": "BTC",
                    "cost": 0.0009,
                    "rate": 0.002
                },
                "info": {}
            }
            """)

        order = create_order_instance_from_raw(trader, raw_order)
        order.tag = "tag"
        trade = create_trade_from_order(order, close_status=OrderStatus.FILLED)

        assert trade.trade_id == '12345-67890:09876/54321'
        assert trade.origin_order_id == '12345-67890:09876/54321'
        assert trade.exchange_order_id == '12345-67890:09876/1111'
        assert trade.exchange_trade_id is None
        assert trade.simulated is True
        assert trade.trade_type == TraderOrderType.SELL_LIMIT
        assert trade.symbol == 'BTC/USDT'
        assert trade.total_cost == decimal.Decimal(str(0.076094524))
        assert trade.executed_quantity == decimal.Decimal(str(1.1))
        assert trade.origin_quantity == decimal.Decimal(str(1.5))
        assert trade.origin_price == decimal.Decimal("7684")
        assert trade.executed_price == decimal.Decimal("7684")
        assert trade.status == OrderStatus.FILLED
        assert trade.is_closing_order is True
        assert trade.tag == 'tag'

        trade = create_trade_from_order(order)
        assert trade.status == OrderStatus.FILLED

        exec_time = time.time()
        trade = create_trade_from_order(order, executed_time=exec_time, exchange_trade_id="plop-trade1")
        assert trade.executed_time == exec_time
        assert trade.exchange_trade_id == "plop-trade1"

        # market order
        raw_order = {
            'id': '362550114',
            'exchange_id': 'AaaaAA',
            'clientOrderId': 'x-T9698eeeeeeeeeeeeee792',
            'timestamp': 1637579281.377,
            'datetime': '2021-11-22T11:08:01.377Z',
            'lastTradeTimestamp': None,
            'symbol': 'UNI/USDT',
            'type': 'market',
            'timeInForce': 'GTC',
            'postOnly': False,
            'side': 'sell',
            'price': None,
            'stopPrice': None,
            'amount': 44964.0,
            'cost': None,
            'average': None,
            'filled': 44964.0,
            'remaining': 0.0,
            'status': 'closed',
            'fee': {'cost': 0.03764836, 'currency': 'USDT'},
            'trades': [],
            'fees': []
        }
        order = create_order_instance_from_raw(trader, raw_order)
        trade = create_trade_from_order(order, close_status=OrderStatus.FILLED, exchange_trade_id="trade_exchange")

        assert trade.trade_id == '362550114'
        assert trade.origin_order_id == '362550114'
        assert trade.exchange_order_id == 'AaaaAA'
        assert trade.exchange_trade_id == 'trade_exchange'
        assert trade.trade_type == TraderOrderType.SELL_MARKET
        assert trade.symbol == 'UNI/USDT'
        assert trade.total_cost == ZERO
        assert trade.executed_quantity == decimal.Decimal("44964.0")
        assert trade.origin_quantity == decimal.Decimal("44964.0")
        assert trade.origin_price == ZERO
        assert trade.executed_price == ZERO
        assert trade.status == OrderStatus.FILLED
        assert trade.executed_time == 1637579281.377
        assert trade.is_closing_order is True

        await self.stop(exchange_manager)

    async def test_create_trade_from_partially_filled_order(self):
        _, exchange_manager, trader = await self.init_default()

        raw_order = json.loads(
            """
            {
                "exchange_id":       "plopplip",
                "datetime":          "2017-08-17 12:42:48.000",
                "timestamp":          1502962946216,
                "lastTradeTimestamp": 1502962956216,
                "status":     "open",
                "symbol":     "BTC/USDT",
                "type":       "limit",
                "side":       "sell",
                "price":       7684,
                "amount":      1.5,
                "filled":      1.1,
                "remaining":   0.4,
                "cost":        0.076094524,
                "trades":    [],
                "fee": {
                    "currency": "BTC",
                    "cost": 0.0009,
                    "rate": 0.002
                },
                "info": {}
            }
            """)

        order = create_order_instance_from_raw(trader, raw_order)
        trade = create_trade_from_order(order, close_status=OrderStatus.OPEN)

        assert trade.trade_id is not None
        assert trade.origin_order_id == trade.trade_id
        assert trade.exchange_order_id == 'plopplip'
        assert trade.exchange_trade_id is None
        assert trade.simulated is True
        assert trade.trade_type == TraderOrderType.SELL_LIMIT
        assert trade.symbol == 'BTC/USDT'
        assert trade.total_cost == decimal.Decimal(str(0.076094524))
        assert trade.executed_quantity == decimal.Decimal(str(1.1))
        assert trade.origin_quantity == decimal.Decimal(str(1.5))
        assert trade.origin_price == decimal.Decimal(str(7684))
        assert trade.executed_price == decimal.Decimal(str(7684))
        assert trade.status == OrderStatus.OPEN
        assert trade.is_closing_order is False

        await self.stop(exchange_manager)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import math
import typing
import octobot_commons.logging as logging

import octobot_trading.enums as enums
import octobot_trading.util as util


class TickerManager(util.Initializable):
    def __init__(self):
        super().__init__()
        self.logger: logging.BotLogger = logging.get_logger(self.__class__.__name__)
        self.ticker: dict[str, typing.Any] = {}
        self.mini_ticker: dict[str, typing.Any] = {}
        self.reset_ticker()
        self.reset_mini_ticker()

    async def initialize_impl(self):
        self.reset_ticker()
        self.reset_mini_ticker()

    def reset_mini_ticker(self):
        self.mini_ticker = {
            enums.ExchangeConstantsMiniTickerColumns.HIGH_PRICE.value: math.nan,
            enums.ExchangeConstantsMiniTickerColumns.LOW_PRICE.value: math.nan,
            enums.ExchangeConstantsMiniTickerColumns.OPEN_PRICE.value: math.nan,
            enums.ExchangeConstantsMiniTickerColumns.CLOSE_PRICE.value: math.nan,
            enums.ExchangeConstantsMiniTickerColumns.VOLUME.value: math.nan,
            enums.ExchangeConstantsMiniTickerColumns.TIMESTAMP.value: 0
        }

    def reset_ticker(self):
        self.ticker = {
            enums.ExchangeConstantsTickersColumns.ASK.value: math.nan,
            enums.ExchangeConstantsTickersColumns.ASK_VOLUME.value: math.nan,
            enums.ExchangeConstantsTickersColumns.BID.value: math.nan,
            enums.ExchangeConstantsTickersColumns.BID_VOLUME.value: math.nan,
            enums.ExchangeConstantsTickersColumns.OPEN.value: math.nan,
            enums.ExchangeConstantsTickersColumns.LOW.value: math.nan,
            enums.ExchangeConstantsTickersColumns.HIGH.value: math.nan,
            enums.ExchangeConstantsTickersColumns.CLOSE.value: math.nan,
            enums.ExchangeConstantsTickersColumns.LAST.value: math.nan,
            enums.ExchangeConstantsTickersColumns.AVERAGE.value: math.nan,
            enums.ExchangeConstantsTickersColumns.SYMBOL.value: math.nan,
            enums.ExchangeConstantsTickersColumns.BASE_VOLUME.value: math.nan,
            enums.ExchangeConstantsTickersColumns.QUOTE_VOLUME.value: math.nan,
            enums.ExchangeConstantsTickersColumns.TIMESTAMP.value: 0,
            enums.ExchangeConstantsTickersColumns.VWAP.value: math.nan
        }

    def ticker_update(self, ticker):
        ticker_update = {
            key: val
            for key, val in ticker.items()
            if val
        }
        self.ticker.update(ticker_update)

    def mini_ticker_update(self, mini_ticker):
        self.mini_ticker.update(mini_ticker)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal

import octobot_commons.logging as logging

import octobot_trading.enums as enums
import octobot_trading.constants as constants
import octobot_trading.exchange_data.contracts.future_contract as future_contract


def update_contracts_from_positions(exchange_manager, positions) -> bool:
    updated = False
    for position in positions:
        pair = position.get(enums.ExchangeConstantsPositionColumns.SYMBOL.value, None)
        if pair and pair not in exchange_manager.exchange.pair_contracts:
            contract = exchange_manager.exchange.create_pair_contract(
                pair=pair,
                current_leverage=position.get(
                    enums.ExchangeConstantsPositionColumns.LEVERAGE.value, constants.ZERO),
                contract_size=position.get(
                    enums.ExchangeConstantsPositionColumns.CONTRACT_SIZE.value, constants.DEFAULT_SYMBOL_CONTRACT_SIZE),
                margin_type=position.get(enums.ExchangeConstantsPositionColumns.MARGIN_TYPE.value, None),
                contract_type=position.get(enums.ExchangeConstantsPositionColumns.CONTRACT_TYPE.value, None),
                position_mode=position.get(enums.ExchangeConstantsPositionColumns.POSITION_MODE.value, None),
                maintenance_margin_rate=position.get(
                    enums.ExchangeConstantsPositionColumns.MAINTENANCE_MARGIN_RATE.value,
                    constants.DEFAULT_SYMBOL_MAINTENANCE_MARGIN_RATE)
            )
            # contracts are updated from positions are fully initialized
            if contract.is_handled_contract():
                updated = True
            else:
                message = f"Unhandled contract {contract}. This contract can't be traded"
                if pair in exchange_manager.exchange_config.traded_symbol_pairs:
                    # inform user that the contract can't be used
                    _get_logger().error(message)
                else:
                    # no need to inform as the contract is not requested
                    _get_logger().debug(message)
    return updated


def update_future_contract_from_dict(exchange_manager, contract: dict) -> bool:
    return exchange_manager.exchange.create_pair_contract(
        pair=contract[enums.ExchangeConstantsMarginContractColumns.PAIR.value],
        current_leverage=decimal.Decimal(str(
            contract[enums.ExchangeConstantsMarginContractColumns.CURRENT_LEVERAGE.value]
        )),
        contract_size=decimal.Decimal(str(
            contract[enums.ExchangeConstantsMarginContractColumns.CONTRACT_SIZE.value]
        )),
        margin_type=enums.MarginType(contract[enums.ExchangeConstantsMarginContractColumns.MARGIN_TYPE.value]),
        contract_type=enums.FutureContractType(contract[enums.ExchangeConstantsFutureContractColumns.CONTRACT_TYPE.value]),
        position_mode=enums.PositionMode(contract[enums.ExchangeConstantsFutureContractColumns.POSITION_MODE.value])
            if contract[enums.ExchangeConstantsFutureContractColumns.POSITION_MODE.value]
            else contract[enums.ExchangeConstantsFutureContractColumns.POSITION_MODE.value],
        maintenance_margin_rate=decimal.Decimal(str(
            contract[enums.ExchangeConstantsFutureContractColumns.MAINTENANCE_MARGIN_RATE.value]
        )),
        maximum_leverage=None if contract[enums.ExchangeConstantsMarginContractColumns.MAXIMUM_LEVERAGE.value] is None
            else decimal.Decimal(str(
                contract[enums.ExchangeConstantsMarginContractColumns.MAXIMUM_LEVERAGE.value]
            ))
    )


def create_default_future_contract(
    pair: str, leverage: decimal.Decimal, contract_type: enums.FutureContractType, position_mode: enums.PositionMode
) -> future_contract.FutureContract:
    return future_contract.FutureContract(
        pair=pair,
        contract_size=constants.DEFAULT_SYMBOL_CONTRACT_SIZE,
        margin_type=constants.DEFAULT_SYMBOL_MARGIN_TYPE,
        contract_type=contract_type,
        maximum_leverage=constants.DEFAULT_SYMBOL_MAX_LEVERAGE,
        current_leverage=leverage,
        position_mode=position_mode,
        maintenance_margin_rate=constants.DEFAULT_SYMBOL_MAINTENANCE_MARGIN_RATE
    )


def _get_logger():
    return logging.get_logger("contract_factory")



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.enums as enums
import octobot_trading.personal_data.orders.types.limit.limit_order as limit_order


class TakeProfitOrder(limit_order.LimitOrder):
    def __init__(self, trader, side=enums.TradeOrderSide.SELL):
        super().__init__(trader, side)
        self.trigger_above = self.side is enums.TradeOrderSide.SELL

    async def update_price_if_outdated(self):
        pass

    def _filled_maker_or_taker(self):
        # Creates a market order when filled, which is taker
        return enums.ExchangeConstantsMarketPropertyColumns.TAKER.value

    async def on_filled(self, enable_associated_orders_creation, force_artificial_orders=False):
        await limit_order.LimitOrder.on_filled(
            self, enable_associated_orders_creation, force_artificial_orders=force_artificial_orders
        )
        if not self.trader.simulate and (self.is_self_managed() or force_artificial_orders):
            order_type = (
                enums.TraderOrderType.SELL_MARKET
                if self.side is enums.TradeOrderSide.SELL else enums.TraderOrderType.BUY_MARKET
            )
            self.on_filled_artificial_order = await self.trader.create_artificial_order(
                order_type, self.symbol, self.origin_stop_price, self.origin_quantity, self.origin_stop_price,
                self.reduce_only, self.close_position
            )



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import ccxt

import octobot_commons.logging as commons_logging
import octobot_commons.html_util as html_util
import octobot_trading.constants as constants
import octobot_trading.errors as errors


def retried_failed_network_request(
    attempts: int = constants.FAILED_NETWORK_REQUEST_RETRY_ATTEMPTS,
    retriable_proxy_errors_attempts: int = constants.FAILED_PROXY_NETWORK_REQUEST_RETRY_ATTEMPTS,
    delay: float = constants.DEFAULT_FAILED_REQUEST_RETRY_TIME,
):
    # inner level to allow passing params to the decorator
    def inner_retried_failed_network_request(func):
        async def _retried_failed_network_request_wrapper(*args, **kwargs):
            proxy_error_attempts = 0
            for attempt in range(attempts):
                retriable_error = False
                try:
                    resp = await func(*args, **kwargs)
                    if attempt > 0:
                        commons_logging.get_logger(f"retried_failed_network_request").info(
                            f"{func.__name__} succeeded after {attempt+1} attempts."
                        )
                    return resp
                except (
                    ccxt.RequestTimeout, ccxt.ExchangeNotAvailable, ccxt.InvalidNonce,
                    errors.RetriableFailedRequest, errors.RetriableExchangeProxyError
                ) as err:
                    retriable_error = err
                    if isinstance(err, errors.RetriableExchangeProxyError):
                        if proxy_error_attempts >= retriable_proxy_errors_attempts - 1:
                            commons_logging.get_logger(f"retried_failed_network_request").warning(
                                f"{func.__name__} raised {html_util.get_html_summary_if_relevant(err)} "
                                f"({err.__class__.__name__}) [proxy error attempts "
                                f"{proxy_error_attempts+1}/{retriable_proxy_errors_attempts}]. "
                                f"Aborting retry attempts."
                            )
                            raise
                        proxy_error_attempts += 1
                    if attempt >= attempts - 1:
                        # no more attempt remaining
                        raise
                except ccxt.ExchangeError as err:
                    retriable_error = err
                    is_retriable_error = False
                    for error_message in constants.RETRIABLE_EXCHANGE_ERRORS_DESC:
                        if error_message in str(err):
                            is_retriable_error = True
                    if attempt >= attempts - 1 or not is_retriable_error:
                        raise
                # propagate any other error

                # did not propagate error, retry after delay
                commons_logging.get_logger(f"retried_failed_network_request").warning(
                    f"{func.__name__} raised {html_util.get_html_summary_if_relevant(retriable_error)} "
                    f"({retriable_error.__class__.__name__}) [attempts {attempt+1}/{attempts}]. Retrying in {delay} seconds."
                )
                await asyncio.sleep(delay)
            return None
        return _retried_failed_network_request_wrapper
    return inner_retried_failed_network_request



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import mock
import pytest

import octobot_trading.enums
import octobot_trading.personal_data as personal_data
from tests.exchanges import backtesting_trader, backtesting_config, backtesting_exchange_manager, fake_backtesting

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_initial_portfolio(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager

    assert isinstance(portfolio_manager.portfolio, personal_data.SpotPortfolio)


async def test_update_portfolio_data_from_order(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager
    pass


async def test_update_portfolio_available_from_order(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager

    # Test buy order
    market_buy = personal_data.BuyMarketOrder(trader)
    market_buy.update(order_type=octobot_trading.enums.TraderOrderType.BUY_MARKET,
                      symbol="BTC/USDT",
                      current_price=70,
                      quantity=10,
                      price=70)

    # test buy order creation
    portfolio_manager.portfolio.update_portfolio_available(market_buy, True)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('10')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('300')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('10')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000')

    # test buy order canceled --> return to init state and the update_portfolio will sync TOTAL with AVAILABLE
    portfolio_manager.portfolio.update_portfolio_available(market_buy, False)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('10')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('1000')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('10')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000')

    # Test sell order
    limit_sell = personal_data.SellLimitOrder(trader)
    limit_sell.update(order_type=octobot_trading.enums.TraderOrderType.SELL_LIMIT,
                      symbol="BTC/USDT",
                      current_price=60,
                      quantity=8,
                      price=60)

    # test sell order creation
    portfolio_manager.portfolio.update_portfolio_available(limit_sell, True)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('2')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('1000')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('10')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000')

    # test sell order canceled --> return to init state and the update_portfolio will sync TOTAL with AVAILABLE
    portfolio_manager.portfolio.update_portfolio_available(limit_sell, False)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('10')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('1000')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('10')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000')

    # Test buy order with large USDT fee
    limit_buy = personal_data.BuyLimitOrder(trader)
    limit_buy.update(order_type=octobot_trading.enums.TraderOrderType.BUY_LIMIT,
                      symbol="BTC/USDT",
                      current_price=60,
                      quantity=8,
                      quantity_filled=8,
                      price=60)
    fee = {
        octobot_trading.enums.FeePropertyColumns.IS_FROM_EXCHANGE.value: True,
        octobot_trading.enums.FeePropertyColumns.COST.value: 54.111111,
        octobot_trading.enums.FeePropertyColumns.CURRENCY.value: "USDT",
    }

    # even if order is filled, always update available funds with use_origin_quantity_and_price and taker fees
    with (mock.patch.object(limit_buy, "is_filled", mock.Mock(return_value=True)) as is_filled_value_mock,
          mock.patch.object(limit_buy.exchange_manager.exchange, "get_trade_fee", mock.Mock(return_value=fee)) as get_trade_fee_mock):
        # test buy order creation
        portfolio_manager.portfolio.update_portfolio_available(limit_buy, True)
        get_trade_fee_mock.assert_called_once_with(
            limit_buy.symbol, limit_buy.order_type, limit_buy.origin_quantity, limit_buy.origin_price,
            octobot_trading.enums.ExchangeConstantsOrderColumns.TAKER.value,
        )
        get_trade_fee_mock.reset_mock()
        is_filled_value_mock.assert_called_once()
        is_filled_value_mock.reset_mock()
        assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('10')
        # cost with fees = 60 * 8 + 54.111111 = 534.111111
        assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == (
            decimal.Decimal('1000') - decimal.Decimal('534.111111')
        )
        assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('10')
        assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000')

        # test buy order canceled --> return to init state and the update_portfolio will sync TOTAL with AVAILABLE
        portfolio_manager.portfolio.update_portfolio_available(limit_buy, False)
        get_trade_fee_mock.assert_called_once_with(
            limit_buy.symbol, limit_buy.order_type, limit_buy.origin_quantity, limit_buy.origin_price,
            octobot_trading.enums.ExchangeConstantsOrderColumns.TAKER.value,
        )
        get_trade_fee_mock.reset_mock()
        is_filled_value_mock.assert_called_once()
        is_filled_value_mock.reset_mock()
        assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('10')
        assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('1000')
        assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('10')
        assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000')


async def test_update_portfolio_available_from_partially_filled_orders(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager

    # Test buy order
    limit_buy = personal_data.BuyLimitOrder(trader)
    limit_buy.update(order_type=octobot_trading.enums.TraderOrderType.BUY_LIMIT,
                      symbol="BTC/USDT",
                      current_price=decimal.Decimal('70'),
                      quantity=decimal.Decimal('10'),
                      quantity_filled=decimal.Decimal('3'),
                      price=decimal.Decimal('70'))

    already_filled_cost = limit_buy.get_cost(limit_buy.filled_quantity)
    assert already_filled_cost == decimal.Decimal('210')
    assert limit_buy.get_locked_quantity() == decimal.Decimal('7') # 10 - 3
    # test buy order creation
    portfolio_manager.portfolio.update_portfolio_available(limit_buy, True)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('10')
    # already_filled_cost is not taken from available funds
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('300') + already_filled_cost
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('10')
    # in this test, total funds have not been updated to account for already filled cost, so it's still 1000
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000')

    # test buy order canceled --> return to init state and the update_portfolio will sync TOTAL with AVAILABLE
    portfolio_manager.portfolio.update_portfolio_available(limit_buy, False)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('10')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('1000')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('10')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000')

    # Test sell order
    limit_sell = personal_data.SellLimitOrder(trader)
    limit_sell.update(order_type=octobot_trading.enums.TraderOrderType.SELL_LIMIT,
                      symbol="BTC/USDT",
                      current_price=decimal.Decimal('60'),
                      quantity=decimal.Decimal('8'),
                      quantity_filled=decimal.Decimal('7.5'),
                      price=decimal.Decimal('60'))

    already_filled_cost = limit_sell.get_cost(limit_sell.filled_quantity)
    assert already_filled_cost == decimal.Decimal('450')    # 7.5 * 60
    assert limit_sell.get_locked_quantity() == decimal.Decimal('0.5') # 8 - 7.5

    # test sell order creation
    portfolio_manager.portfolio.update_portfolio_available(limit_sell, True)
    # 0.5 locked as other 7.5 have already been sold, they are not locked anymore
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('10') - decimal.Decimal('0.5')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('1000')
    # in this test, total has not been updated to account for already filled amount, so it's still 10
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('10')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000')

    # test sell order canceled --> return to init state and the update_portfolio will sync TOTAL with AVAILABLE
    portfolio_manager.portfolio.update_portfolio_available(limit_sell, False)
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").available == decimal.Decimal('10')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").available == decimal.Decimal('1000')
    assert portfolio_manager.portfolio.get_currency_portfolio("BTC").total == decimal.Decimal('10')
    assert portfolio_manager.portfolio.get_currency_portfolio("USDT").total == decimal.Decimal('1000')



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import octobot_commons.html_util as html_util

import octobot_trading.enums as enums
import octobot_trading.constants as constants
import octobot_trading.personal_data.orders.order_state as order_state
import octobot_trading.personal_data.orders.states.order_state_factory as order_state_factory


class CancelOrderState(order_state.OrderState):
    MAX_SYNCHRONIZATION_ATTEMPTS = 5

    def __init__(self, order, is_from_exchange_data, enable_associated_orders_creation=True,
        is_already_counted_in_available_funds=False
    ):
        super().__init__(
            order, is_from_exchange_data, enable_associated_orders_creation=enable_associated_orders_creation,
            is_already_counted_in_available_funds=is_already_counted_in_available_funds
        )
        self.state = enums.OrderStates.CANCELING \
            if ((not self.order.simulated and not self.is_status_cancelled()) or self.is_status_pending()) \
            else enums.OrderStates.CANCELED

    async def initialize_impl(self, forced=False, ignored_order=None) -> None:
        if forced:
            self._force_final_state()

        if (
            self.order.order_group and self.enable_associated_orders_creation
            and not self.order.is_in_active_inactive_transition
        ):
            await self.order.order_group.on_cancel(self.order, ignored_orders=[ignored_order])

        await super().initialize_impl()

    def is_pending(self) -> bool:
        return self.state is enums.OrderStates.CANCELING

    def is_canceled(self) -> bool:
        return self.state is enums.OrderStates.CANCELED

    def is_status_pending(self) -> bool:
        return self.order.status is enums.OrderStatus.PENDING_CANCEL and not self.order.simulated

    def is_status_cancelled(self) -> bool:
        return not self.is_status_pending() and self.order.status in constants.CANCEL_ORDER_STATUS_SCOPE

    def _force_final_state(self):
        self.state = enums.OrderStates.CANCELED
        self.order.status = enums.OrderStatus.CANCELED

    def allows_new_status(self, status) -> bool:
        """
        Don't allow going from canceling to open
        :return: True if the given status is compatible with the current state
        """
        return status in constants.CANCEL_ORDER_STATUS_SCOPE or status in constants.FILL_ORDER_STATUS_SCOPE

    async def _synchronize_with_exchange(self, force_synchronization: bool = False) -> None:
        """
        Ask OrdersChannel Internal producer to refresh the order from the exchange
        :param force_synchronization: When True, for the update of the order from the exchange
        :return: the result of OrdersProducer.update_order_from_exchange()
        """
        if not self.has_already_been_synchronized_once:
            # If we want to sync this state, it means the order is being canceled by the exchange but is not
            # fully canceled yet. Giving some time to the exchange before re-requesting it.
            self.get_logger().info(
                f"{self.__class__.__name__} still pending, synchronizing in {self.PENDING_REFRESH_INTERVAL}s"
            )
            await asyncio.sleep(self.PENDING_REFRESH_INTERVAL)
        await super()._synchronize_with_exchange(force_synchronization=force_synchronization)

    async def on_refresh_successful(self):
        """
        Verify the order is properly canceled
        """
        if self.is_status_pending():
            await self.update()
        elif self.order.status in constants.CANCEL_ORDER_STATUS_SCOPE:
            self.state = enums.OrderStates.CANCELED
            await self.update()
        else:
            await order_state_factory.create_order_state(self.order, is_from_exchange_data=True,
                                                         ignore_states=[enums.States.OPEN])

    async def terminate(self):
        """
        Replace the order state by a close state
        `force_close = True` because we know that the order is successfully cancelled.
        """
        try:
            self.log_event_message(enums.StatesMessages.CANCELLED)
            self.ensure_not_cleared(self.order)

            # set cancel time
            self.order.canceled_time = self.order.exchange_manager.exchange.get_exchange_current_time()

            # update portfolio after close
            async with self.order.exchange_manager.exchange_personal_data.portfolio_manager.portfolio.lock:
                self.ensure_not_cleared(self.order)
                await self.order.exchange_manager.exchange_personal_data.\
                    handle_portfolio_and_position_update_from_order(self.order, require_exchange_update=False)

            # notify order cancelled
            await self.order.exchange_manager.exchange_personal_data.handle_order_update_notification(
                self.order, enums.OrderUpdateType.STATE_CHANGE
            )

            # set close state
            await self.order.on_close(force_close=True)  # TODO force ?
        except Exception as e:
            self.get_logger().exception(
                e, True, f"Fail to execute cancel state termination : {html_util.get_html_summary_if_relevant(e)}."
            )
            raise



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import dataclasses
import typing
import octobot_commons.logging
import octobot_trading.personal_data.orders.order_util as order_util
import octobot_trading.enums as enums
import octobot_trading.constants as constants


@dataclasses.dataclass
class ActiveOrderSwapStrategy:
    """
    Will wait up to the given timeout for the non-priority orders to be filled.
    If it is not filled during this time, priority will be re-created and lower priority order will
    return to inactive.
    """
    swap_timeout: float = constants.ACTIVE_ORDER_STRATEGY_SWAP_TIMEOUT
    trigger_price_configuration: str = enums.ActiveOrderSwapTriggerPriceConfiguration.FILLING_PRICE.value

    def is_priority_order(self, order) -> bool:
        raise NotImplementedError("is_priority_order is not implemented")

    async def apply_inactive_orders(self, orders: list):
        for order in orders:
            trigger_price = self._get_trigger_price(order)
            if self.is_priority_order(order):
                # still register active trigger in case this order becomes inactive
                order.update(
                    active_trigger=order_util.create_order_price_trigger(order, trigger_price, order.trigger_above)
                )
            else:
                await order.set_as_inactive(
                    order_util.create_order_price_trigger(order, trigger_price, order.trigger_above)
                )

    def on_order_update(self, order, update_time):
        if order.active_trigger:
            order.active_trigger.update(
                trigger_price=self._get_trigger_price(order), min_trigger_time=update_time,
                update_event=order.is_synchronization_enabled()
            )

    def _get_trigger_price(self, order) -> decimal.Decimal:
        if self.trigger_price_configuration == enums.ActiveOrderSwapTriggerPriceConfiguration.FILLING_PRICE.value:
            return order.get_filling_price()
        if self.trigger_price_configuration == enums.ActiveOrderSwapTriggerPriceConfiguration.ORDER_PARAMS_ONLY.value:
            if order.active_trigger is None or order.active_trigger.trigger_price is None:
                raise ValueError(
                    f"order.active_trigger.trigger_price must be set when using "
                    f"ActiveOrderSwapTriggerPriceConfiguration.ORDER_PARAMS_ONLY. Order: {order}"
                )
            return order.active_trigger.trigger_price
        raise ValueError(f"Unknown trigger price configuration: {self.trigger_price_configuration}")

    async def execute(
        self,
        inactive_order,
        wait_for_fill_callback: typing.Optional[typing.Callable],
        timeout: typing.Optional[float]
    ):
        if inactive_order.order_group is None:
            raise NotImplementedError(f"Input order is not part of a group, this is unexpected: {inactive_order}")
        timeout = self.swap_timeout if timeout is None else timeout
        # strategies should not be executed concurrently or in parallel with other group triggers
        async with inactive_order.order_group.lock_group():
            active_order, now_maybe_partially_inactive_orders, reverse_update_callback = (
                await self._update_group_and_activate_order(inactive_order)
            )
            if active_order is None:
                raise ValueError("No active order was created")
            if not any(self.is_priority_order(inactive_order) for inactive_order in now_maybe_partially_inactive_orders):
                # nothing else to do: no priority order has been deactivated
                return
            # priority order have been deactivated: if newly active order doesn't get filled within timeout,
            # to should get back to inactivity and priority orders should get back to active
            logger = octobot_commons.logging.get_logger(active_order.get_logger_name())
            logger.info(f"Waiting for newly active order to be filled timeout={timeout}s")
            await (
                wait_for_fill_callback(active_order, timeout) if wait_for_fill_callback
                else order_util.wait_for_order_fill(active_order, timeout, True)
            )
            if active_order.is_open() and not (active_order.is_filled() or active_order.is_closed()):
                logger.warning(
                    f"Newly active order was not filled within {timeout}, reversing active "
                    f"order to re-activate priority orders."
                )
                await reverse_update_callback(active_order, now_maybe_partially_inactive_orders)

    async def _update_group_and_activate_order(self, inactive_order) -> (list, typing.Callable[[list], None]):
        if inactive_order.is_active is True:
            raise ValueError(f"Order is active already: {inactive_order}")
        # 1. cancel/edit the other order(s) that need to be canceled first and then create this order
        now_maybe_partially_inactive_orders, reverse_update_callback = \
            await inactive_order.order_group.adapt_before_order_becoming_active(inactive_order)
        # 2. update this order as active
        active_order = await order_util.create_as_active_order_on_exchange(inactive_order, False)
        return active_order, now_maybe_partially_inactive_orders, reverse_update_callback



# pylint: disable=E0611
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio

import octobot_commons.async_job as async_job
import octobot_commons.html_util as html_util

import octobot_trading.errors as errors
import octobot_trading.enums as enums
import octobot_trading.personal_data.positions.channel.positions as positions_channel
import octobot_trading.constants as constants
import octobot_trading.exchange_channel as exchanges_channel
import octobot_trading.exchange_data as exchange_data


class PositionsUpdater(positions_channel.PositionsProducer):
    """
    Update positions from exchange
    Can also be used to update a specific positions from exchange
    """

    CHANNEL_NAME = constants.POSITIONS_CHANNEL
    POSITIONS_STARTING_REFRESH_TIME = 12
    POSITION_REFRESH_TIME = 9
    TIME_BETWEEN_POSITIONS_REFRESH = 3

    def __init__(self, channel):
        super().__init__(channel)
        # create async jobs
        self.positions_update_job = async_job.AsyncJob(
            self.fetch_and_push_positions,
            execution_interval_delay=self.POSITION_REFRESH_TIME,
            min_execution_delay=self.TIME_BETWEEN_POSITIONS_REFRESH
        )
        self.position_update_job = async_job.AsyncJob(
            self._position_fetch_and_push,
            is_periodic=False,
            enable_multiple_runs=True
        )
        self.position_update_job.add_job_dependency(self.positions_update_job)
        self.positions_update_job.add_job_dependency(self.position_update_job)

    async def initialize(self) -> None:
        """
        Initialize positions and future contracts
        """
        # fetch future contracts from exchange
        await self.initialize_contracts()

        # subscribe to mark_price channel if necessary
        if not self._has_mark_price_in_position():
            await exchanges_channel.get_chan(constants.MARK_PRICE_CHANNEL, self.channel.exchange_manager.id) \
                .new_consumer(self.handle_mark_price)

        # fetch current positions from exchange
        await self.initialize_positions()

        for pair in self.channel.exchange_manager.exchange_config.traded_symbol_pairs:
            self.channel.exchange_manager.exchange_personal_data.positions_manager.set_initialized_event(pair)

    async def initialize_contracts(self) -> None:
        """
        Initialize exchange FutureContracts required to manage positions
        """
        for pair in self.channel.exchange_manager.exchange_config.traded_symbol_pairs:
            try:
                await self.channel.exchange_manager.exchange.load_pair_future_contract(pair)
                await self._update_contract_settings(pair)
            except NotImplementedError as e:
                self.logger.debug(f"Can't to load {pair} contract info from exchange: {e}. "
                                  f"This contract will be created from fetched positions.")
            except Exception as e:
                self.logger.exception(e, False)
                self.logger.warning(
                    f"Failed to load {pair} contract info : {html_util.get_html_summary_if_relevant(e)}"
                )

    async def initialize_positions(self) -> None:
        """
        Initialize data before starting jobs
        """
        try:
            await self.fetch_and_push()
        except NotImplementedError:
            self.logger.warning("Position updater cannot fetch positions : required methods are not implemented")
            await self.stop()
        except errors.NotSupported:
            self.logger.warning(f"{self.channel.exchange_manager.exchange_name} is not supporting updates")
            await self.stop()
        except Exception as e:
            self.logger.exception(
                e,
                True,
                f"Fail to initialize positions : {html_util.get_html_summary_if_relevant(e)}"
            )

    async def start(self):
        """
        Start updater jobs
        """
        if not self._should_run():
            return

        await self.initialize()
        await asyncio.sleep(self.POSITIONS_STARTING_REFRESH_TIME)
        await self.positions_update_job.run()

    async def fetch_and_push(self):
        """
        Update positions from exchange
        """
        await self.fetch_and_push_positions()
        await asyncio.sleep(self.TIME_BETWEEN_POSITIONS_REFRESH)

    def _should_run(self):
        return self.channel.exchange_manager.is_future

    def _is_relevant_position(self, position_dict):
        return position_dict and position_dict.get(enums.ExchangeConstantsPositionColumns.SYMBOL.value, None) \
               in self.channel.exchange_manager.exchange_config.traded_symbol_pairs

    async def fetch_and_push_positions(self, retry_attempts=1):
        """
        Update positions from exchange
        """
        symbols = self.channel.exchange_manager.exchange_config.traded_symbol_pairs \
            if self.channel.exchange_manager.exchange.REQUIRES_SYMBOL_FOR_EMPTY_POSITION else None
        positions = await self.channel.exchange_manager.exchange.retry_n_time(
            retry_attempts,
            self.channel.exchange_manager.exchange.get_positions, symbols=symbols,
        )
        if positions:
            relevant_positions = [
                position
                for position in positions
                if self._is_relevant_position(position)
            ]
            # initialize relevant contracts first as they might be waited for
            updated = exchange_data.update_contracts_from_positions(self.channel.exchange_manager, relevant_positions)
            if exchange_data.update_contracts_from_positions(self.channel.exchange_manager, positions) or updated:
                await self._update_positions_contract_settings(positions)
            # only consider positions that are relevant to the current setup
            await self._push_positions(relevant_positions)

    async def _push_positions(self, positions):
        await self.push(positions)

        if self._should_push_mark_price():
            for position in positions:
                await self.extract_mark_price(position)

    async def _update_positions_contract_settings(self, positions):
        for position in positions:
            symbol = position.get(enums.ExchangeConstantsPositionColumns.SYMBOL.value, None)
            if symbol is not None and symbol in self.channel.exchange_manager.exchange_config.traded_symbol_pairs:
                await self._update_contract_settings(symbol)

    async def _update_contract_settings(self, symbol):
        try:
            if constants.FORCED_MARGIN_TYPE:
                await self.channel.exchange_manager.trader.set_margin_type(
                    symbol, enums.PositionSide.BOTH, constants.FORCED_MARGIN_TYPE
                )
        except Exception as e:
            self.logger.exception(
                e,
                True,
                f"Fail to update contracts settings : {html_util.get_html_summary_if_relevant(e)}"
            )
        finally:
            self.channel.exchange_manager.exchange.set_contract_initialized_event(symbol)

    async def extract_mark_price(self, position_dict: dict):
        try:
            await exchanges_channel.get_chan(constants.MARK_PRICE_CHANNEL,
                                             self.channel.exchange_manager.id).get_internal_producer(). \
                push(position_dict[enums.ExchangeConstantsPositionColumns.SYMBOL.value],
                     position_dict[enums.ExchangeConstantsPositionColumns.MARK_PRICE.value])
        except Exception as e:
            self.logger.exception(e, True, f"Fail to update mark price from position : {e}")

    async def update_position_from_exchange(self, position,
                                            should_notify=False,
                                            wait_for_refresh=False,
                                            force_job_execution=False,
                                            create_position_producer_if_missing=True):
        await self.position_update_job.run(force=True, wait_for_task_execution=wait_for_refresh,
                                           ignore_dependencies_check=force_job_execution,
                                           position=position, should_notify=should_notify)

    async def _position_fetch_and_push(self, position, should_notify=False):
        """
        Update Position from exchange
        :param position: the position to update
        :param should_notify: if Positions channel consumers should be notified
        :return: True if the position was updated
        """
        exchange_name = self.channel.exchange_manager.exchange_name
        self.logger.debug(f"Requested update for position: {position} on {exchange_name}")
        raw_position = await self.channel.exchange_manager.exchange.get_position(position.symbol)

        if raw_position:
            self.logger.debug(f"Received update for {position} on {exchange_name}: {raw_position}")

            await self.channel.exchange_manager.exchange_personal_data.handle_position_update(
                symbol=raw_position[enums.ExchangeConstantsPositionColumns.SYMBOL.value],
                side=raw_position[enums.ExchangeConstantsPositionColumns.SIDE.value],
                raw_position=raw_position,
                should_notify=should_notify
            )
        else:
            self.logger.debug(
                f"Can't received update for {position} on {exchange_name}: received position is {raw_position}"
            )


    def _should_push_mark_price(self):
        return self._has_mark_price_in_position()

    def _has_mark_price_in_position(self):
        return self.channel.exchange_manager.exchange.MARK_PRICE_IN_POSITION

    async def handle_mark_price(self, exchange: str, exchange_id: str, cryptocurrency: str, symbol: str, mark_price):
        """
        MarkPrice channel consumer callback
        """
        try:
            for symbol_position in self.channel.exchange_manager.exchange_personal_data.positions_manager. \
                    get_symbol_positions(symbol=symbol):
                await symbol_position.update(mark_price=mark_price)
        except Exception as e:
            self.logger.exception(e, True, f"Fail to handle mark price : {e}")

    async def stop(self) -> None:
        """
        Stop producer by stopping its jobs
        """
        await super().stop()
        if not self._should_run():
            return
        self.positions_update_job.stop()
        self.position_update_job.stop()

    async def resume(self) -> None:
        """
        Resume producer by restarting its jobs
        """
        await super().resume()
        if not self._should_run():
            return
        if not self.is_running:
            await self.run()



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import abc

import octobot_commons.logging as logging
import octobot_commons.tentacles_management as abstract_tentacle
import octobot_tentacles_manager.api as tentacles_manager_api


class AbstractSupervisor(abstract_tentacle.AbstractTentacle):
    __metaclass__ = abc.ABCMeta

    def __init__(self, config, exchange_manager):
        super().__init__()
        self.logger = logging.get_logger(self.get_name())

        # Global OctoBot configuration
        self.config: dict = config

        # Mode related exchange manager instance
        self.exchange_manager = exchange_manager

        # The id of the OctoBot using this trading mode
        self.bot_id: str = None

        # Supervisor specific config (Is loaded from tentacle specific file)
        self.supervisor_config: dict = None

        # If this supervisor is enabled
        self.enabled: bool = True

        # producers is the list of producers created by this supervisor instance
        self.producers = []

        # producers is the list of consumers created by this supervisor instance
        self.consumers = []

    @classmethod
    def get_parent_supervisor_classes(cls, higher_parent_class_limit=None) -> list:
        return [
            class_type
            for class_type in cls.mro()
            if (higher_parent_class_limit if higher_parent_class_limit else AbstractSupervisor) in class_type.mro()
        ]

    @staticmethod
    def is_backtestable() -> bool:
        """
        Should be overwritten
        :return: True if the Supervisor can be used in a backtesting else False
        """
        return True

    async def initialize(self) -> None:
        """
        Initialize supervisor
        """

    async def stop(self) -> None:
        """
        Stops all producers and consumers and clean references
        """
        for producer in self.producers:
            await producer.stop()
        for consumer in self.consumers:
            await consumer.stop()
        self.exchange_manager = None

    def load_config(self) -> None:
        """
        Try to load Supervisor tentacle config.
        Calls set_default_config() if the tentacle config is empty
        """
        # try with this class name
        self.supervisor_config = tentacles_manager_api.get_tentacle_config(self.exchange_manager.tentacles_setup_config,
                                                                           self.__class__)

        # set default config if nothing found
        if not self.supervisor_config:
            self.set_default_config()

    # to implement in subclasses if config is necessary
    def set_default_config(self) -> None:
        pass



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.html_util as html_util
import octobot_trading.enums as enums
import octobot_trading.personal_data.orders.order_state as order_state


class CloseOrderState(order_state.OrderState):
    def __init__(self, order, is_from_exchange_data, enable_associated_orders_creation=True,
        is_already_counted_in_available_funds=False, force_close=True
    ):
        super().__init__(
            order, is_from_exchange_data, enable_associated_orders_creation=enable_associated_orders_creation,
            is_already_counted_in_available_funds=is_already_counted_in_available_funds
        )
        self.state = enums.States.CLOSED if is_from_exchange_data or force_close or self.order.simulated \
            else enums.States.CLOSING

    async def initialize_impl(self, forced=False) -> None:
        if forced:
            self.state = enums.States.CLOSED
        return await super().initialize_impl()

    def is_pending(self) -> bool:
        return self.state is enums.States.CLOSING

    def is_closed(self) -> bool:
        return self.state is enums.States.CLOSED

    async def on_refresh_successful(self):
        """
        Verify the order is properly closed
        """
        if self.order is None or self.order.is_cleared():
            self.get_logger().debug("Skipping on_refresh_successful as order is cleared")
            return
        if self.order.status is enums.OrderStatus.CLOSED:
            self.state = enums.States.CLOSED
            await self.update()

    async def terminate(self):
        """
        Handle order to trade conversion
        """
        try:
            if self.order.is_in_active_inactive_transition:
                # orders that are in active/inactive transition should not generate trades nor be removed from open_orders
                self.get_logger().info(
                    f"Skipping {enums.StatesMessages.CLOSED} state terminate for {self.order} as "
                    f"it is in active/inactive transition"
                )
                await self.order.on_inactive_from_active()
            else:
                self.log_event_message(enums.StatesMessages.CLOSED)

                # add to trade history and notify
                self.ensure_not_cleared(self.order)
                await self.order.exchange_manager.exchange_personal_data.handle_trade_instance_update(
                    self.order.trader.convert_order_to_trade(self.order)
                )

                # remove order from open_orders
                self.order.exchange_manager.exchange_personal_data.orders_manager.remove_order_instance(self.order)
        except Exception as e:
            self.get_logger().exception(
                e, True, f"Fail to execute close state termination : {html_util.get_html_summary_if_relevant(e)}."
            )
            raise

    async def _synchronize_with_exchange(self, force_synchronization=False):
        # Nothing to synchronize
        pass



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import mock
import pytest
import decimal

import octobot_commons.signals as commons_signals
import octobot_commons.asyncio_tools as asyncio_tools

import octobot_trading.constants as constants
import octobot_trading.enums as trading_enums
import octobot_trading.personal_data as trading_personal_data
import octobot_trading.modes.modes_util as modes_util
import octobot_trading.signals as signals

from tests.exchanges import backtesting_trader, backtesting_config, backtesting_exchange_manager, fake_backtesting
from tests import event_loop

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_get_assets_requiring_extra_price_data_to_convert(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    assert modes_util.get_assets_requiring_extra_price_data_to_convert(exchange_manager, [], "USDT") == set()
    assert modes_util.get_assets_requiring_extra_price_data_to_convert(exchange_manager, ["BTC", "ETH"], "USDT") == {
        "BTC"
    }
    converter = exchange_manager.exchange_personal_data.portfolio_manager.portfolio_value_holder.value_converter
    converter.update_last_price("BTC/USDT", decimal.Decimal(1))
    # no ETH in portfolio
    assert modes_util.get_assets_requiring_extra_price_data_to_convert(exchange_manager, ["BTC", "ETH"], "USDT") \
           == set()
    # same when no backtesting
    exchange_manager.is_backtesting = False
    assert modes_util.get_assets_requiring_extra_price_data_to_convert(exchange_manager, ["BTC", "ETH"], "USDT") \
           == set()
    exchange_manager.exchange_personal_data.portfolio_manager.portfolio.portfolio["ETH"] = \
        trading_personal_data.Asset("ETH", constants.ONE, constants.ONE)
    exchange_manager.exchange_personal_data.portfolio_manager.portfolio.portfolio["SOL"] = \
        trading_personal_data.Asset("SOL", constants.ONE, constants.ONE)
    # ETH in portfolio
    assert sorted(list(modes_util.get_assets_requiring_extra_price_data_to_convert(
        exchange_manager, ["BTC", "ETH", "SOL"], "USDT"
    ))) == sorted(list({'ETH', 'SOL'}))
    converter.update_last_price("BTC/ETH", decimal.Decimal(2))
    # can bridge ETH price using BTC
    assert modes_util.get_assets_requiring_extra_price_data_to_convert(exchange_manager, ["BTC", "ETH"], "USDT") \
           == set()
    exchange_manager.is_backtesting = True
    assert modes_util.get_assets_requiring_extra_price_data_to_convert(exchange_manager, ["BTC", "ETH"], "USDT") \
           == set()


async def test_convert_assets_to_target_asset():
    trading_mode = "trading_mode"
    sellable_assets = ["USDT", "PLOP"]
    target_asset = "USD"
    tickers = {"BTC/USDT": {}}
    with mock.patch.object(
            modes_util, "convert_asset_to_target_asset", mock.AsyncMock(return_value=["orders"])
    ) as convert_asset_to_target_asset:
        orders = \
            await modes_util.convert_assets_to_target_asset(trading_mode, sellable_assets, target_asset, tickers)

        assert convert_asset_to_target_asset.call_count == 2
        assert convert_asset_to_target_asset.mock_calls[0].args == \
               (trading_mode, "PLOP", target_asset, {"BTC/USDT": {}})
        assert convert_asset_to_target_asset.mock_calls[0].kwargs == {"asset_amount": None, "dependencies": None}
        assert convert_asset_to_target_asset.mock_calls[1].args == \
               (trading_mode, "USDT", target_asset, {"BTC/USDT": {}})
        assert convert_asset_to_target_asset.mock_calls[1].kwargs == {"asset_amount": None, "dependencies": None}

        assert orders == ["orders", "orders"]

        convert_asset_to_target_asset.reset_mock()
        orders = await modes_util.convert_assets_to_target_asset(trading_mode, [], target_asset, {"tickers": {}})
        convert_asset_to_target_asset.assert_not_called()
        assert orders == []

    with mock.patch.object(
        modes_util, "convert_asset_to_target_asset", mock.AsyncMock(side_effect=KeyError)
    ) as convert_asset_to_target_asset:
        trading_mode = mock.Mock()
        assert await modes_util.convert_assets_to_target_asset(
            trading_mode, sellable_assets, target_asset, tickers,
            dependencies=signals.get_orders_dependencies([
                mock.Mock(order_id="123"),
                mock.Mock(order_id="456")
            ])
        ) == []
        assert sellable_assets
        assert convert_asset_to_target_asset.call_count == len(sellable_assets)
        for call in convert_asset_to_target_asset.mock_calls:
            assert call.kwargs == {"asset_amount": None, "dependencies": signals.get_orders_dependencies([
                mock.Mock(order_id="123"),
                mock.Mock(order_id="456")
            ])}


async def test_convert_asset_to_target_asset(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    trading_mode = _get_trading_mode(exchange_manager)
    sellable_assets = ["USDT", "PLOP"]
    target_asset = "USD"
    tickers = {}
    portfolio = exchange_manager.exchange_personal_data.portfolio_manager.portfolio.portfolio
    converter = exchange_manager.exchange_personal_data.portfolio_manager.portfolio_value_holder.value_converter

    with mock.patch.object(exchange_manager.exchange, "is_market_open_for_order_type", mock.Mock(return_value=True)) \
        as is_market_open_for_order_type_mock:
        # only BTC and USDT in portfolio
        # convert ETH to USDT: nothing to do
        orders = await modes_util.convert_asset_to_target_asset(trading_mode, "ETH", "USDT", tickers)
        assert orders == []
        trading_mode.create_order.assert_not_called()
        is_market_open_for_order_type_mock.assert_not_called()

        # sell BTC
        orders = await modes_util.convert_asset_to_target_asset(trading_mode, "BTC", "USDT", tickers)
        # not working: no BTC pair in exchange_manager.client_symbols (can't create convert order)
        assert orders == []
        trading_mode.create_order.assert_not_called()
        # not called as symbol is not in client_symbols and might not be in symbol_markets
        is_market_open_for_order_type_mock.assert_not_called()

        # add client_symbols pair
        exchange_manager.client_symbols.append("BTC/USDT")

        # not working: no btc price
        assert orders == []
        trading_mode.create_order.assert_not_called()

        # register BTC price
        converter.update_last_price("BTC/USDT", decimal.Decimal(30000))
        orders = await modes_util.convert_asset_to_target_asset(
            trading_mode, "BTC", "USDT", tickers,
            dependencies=signals.get_orders_dependencies([
                mock.Mock(order_id="123"),
                mock.Mock(order_id="456")
            ])
        )
        assert len(orders) == 1
        order = orders[0]
        assert order.order_type == trading_enums.TraderOrderType.SELL_MARKET
        assert order.symbol == "BTC/USDT"
        assert order.origin_quantity == decimal.Decimal(10)
        assert order.created_last_price == decimal.Decimal(30000)
        # market order is instantly filled
        assert order.is_filled()
        assert order.filled_price == order.origin_price
        assert order.filled_quantity > constants.ZERO
        trading_mode.create_order.assert_called_once()
        assert trading_mode.create_order.mock_calls[0].kwargs["dependencies"] == signals.get_orders_dependencies([
            mock.Mock(order_id="123"),
            mock.Mock(order_id="456")
        ])
        trading_mode.create_order.reset_mock()
        is_market_open_for_order_type_mock.assert_called_once_with(
            "BTC/USDT", trading_enums.TraderOrderType.SELL_MARKET
        )
        is_market_open_for_order_type_mock.reset_mock()

    orders = await modes_util.convert_asset_to_target_asset(trading_mode, "USDT", "BTC", tickers)
    assert len(orders) == 1
    order = orders[0]
    assert order.order_type == trading_enums.TraderOrderType.BUY_MARKET
    assert order.symbol == "BTC/USDT"
    assert order.origin_quantity == decimal.Decimal("10.03333333")
    assert order.created_last_price == decimal.Decimal(30000)
    assert order.is_filled()
    trading_mode.create_order.assert_called_once()
    trading_mode.create_order.reset_mock()

    exchange_manager.client_symbols.append("ETH/USDT")
    # using ticker price
    tickers["ETH/USDT"] = {
        trading_enums.ExchangeConstantsTickersColumns.CLOSE.value: decimal.Decimal(1500)
    }

    # reset portfolio
    portfolio["USDT"] = trading_personal_data.SpotAsset("USDT", decimal.Decimal(1000), decimal.Decimal(1000))
    orders = await modes_util.convert_asset_to_target_asset(trading_mode, "USDT", "ETH", tickers)
    assert len(orders) == 1
    order = orders[0]
    assert order.order_type == trading_enums.TraderOrderType.BUY_MARKET
    assert order.symbol == "ETH/USDT"
    assert order.origin_quantity == decimal.Decimal("0.66666666")
    assert order.created_last_price == decimal.Decimal(1500)
    assert order.is_filled()
    trading_mode.create_order.assert_called_once()
    trading_mode.create_order.reset_mock()

    portfolio["ETH"] = trading_personal_data.SpotAsset("ETH", constants.ZERO, constants.ZERO)
    orders = await modes_util.convert_asset_to_target_asset(trading_mode, "ETH", "USDT", tickers)
    # no ETH in portfolio
    assert orders == []
    trading_mode.create_order.assert_not_called()

    portfolio["ETH"] = trading_personal_data.SpotAsset("ETH", constants.ONE, constants.ONE)
    orders = await modes_util.convert_asset_to_target_asset(trading_mode, "ETH", "USDT", tickers)
    assert len(orders) == 1
    order = orders[0]
    assert order.order_type == trading_enums.TraderOrderType.SELL_MARKET
    assert order.symbol == "ETH/USDT"
    assert order.origin_quantity == decimal.Decimal("1")
    assert order.created_last_price == decimal.Decimal(1500)
    assert order.is_filled()
    trading_mode.create_order.assert_called_once()
    trading_mode.create_order.reset_mock()

    # with amount param
    portfolio["ETH"] = trading_personal_data.SpotAsset("ETH", decimal.Decimal("2"), decimal.Decimal("2"))
    orders = await modes_util.convert_asset_to_target_asset(
        trading_mode, "ETH", "USDT", tickers, asset_amount=decimal.Decimal(2)
    )
    assert len(orders) == 1
    order = orders[0]
    assert order.order_type == trading_enums.TraderOrderType.SELL_MARKET
    assert order.symbol == "ETH/USDT"
    assert order.origin_quantity == decimal.Decimal(2)
    assert order.created_last_price == decimal.Decimal(1500)
    assert order.is_filled()
    trading_mode.create_order.assert_called_once()
    trading_mode.create_order.reset_mock()

    orders = await modes_util.convert_asset_to_target_asset(
        trading_mode, "USDT", "ETH", tickers, asset_amount=decimal.Decimal(450)
    )
    assert len(orders) == 1
    order = orders[0]
    assert order.order_type == trading_enums.TraderOrderType.BUY_MARKET
    assert order.symbol == "ETH/USDT"
    assert order.origin_quantity == decimal.Decimal("0.3")
    assert order.created_last_price == decimal.Decimal(1500)
    assert order.is_filled()
    trading_mode.create_order.assert_called_once()
    trading_mode.create_order.reset_mock()

    orders = await modes_util.convert_asset_to_target_asset(
        trading_mode, "USDT", "ETH", tickers, asset_amount=decimal.Decimal(1000)
    )
    assert len(orders) == 1
    order = orders[0]
    assert order.order_type == trading_enums.TraderOrderType.BUY_MARKET
    assert order.symbol == "ETH/USDT"
    assert order.origin_quantity == decimal.Decimal("0.66666666")
    assert order.created_last_price == decimal.Decimal(1500)
    assert order.is_filled()
    trading_mode.create_order.assert_called_once()
    trading_mode.create_order.reset_mock()

    # with fees paid in quote
    fees = {
        trading_enums.FeePropertyColumns.COST.value: "2",
        trading_enums.FeePropertyColumns.CURRENCY.value: "USDT",
        trading_enums.FeePropertyColumns.IS_FROM_EXCHANGE.value: True,
    }
    with mock.patch.object(exchange_manager.exchange, "get_trade_fee", mock.Mock(return_value=fees)) \
         as get_trade_fee_mock:
        # cast 1: enough funds in pf to cover fees
        portfolio["USDT"] = trading_personal_data.SpotAsset("USDT", decimal.Decimal(1000), decimal.Decimal(1000))
        orders = await modes_util.convert_asset_to_target_asset(
            trading_mode, "USDT", "ETH", tickers, asset_amount=decimal.Decimal(450)
        )
        assert len(orders) == 1
        order = orders[0]
        assert order.order_type == trading_enums.TraderOrderType.BUY_MARKET
        assert order.symbol == "ETH/USDT"
        assert order.origin_quantity == decimal.Decimal("0.3")
        assert order.created_last_price == decimal.Decimal(1500)
        assert order.is_filled()
        assert get_trade_fee_mock.call_count == 5
        get_trade_fee_mock.reset_mock()
        trading_mode.create_order.assert_called_once()
        trading_mode.create_order.reset_mock()

        # cast 2: reduce amount to cover fees
        portfolio["USDT"] = trading_personal_data.SpotAsset("USDT", decimal.Decimal(1000), decimal.Decimal(1000))
        orders = await modes_util.convert_asset_to_target_asset(
            trading_mode, "USDT", "ETH", tickers, asset_amount=decimal.Decimal(1000)
        )
        assert len(orders) == 1
        order = orders[0]
        assert order.order_type == trading_enums.TraderOrderType.BUY_MARKET
        assert order.symbol == "ETH/USDT"
        assert order.origin_quantity == decimal.Decimal("0.665")   # lower than 0.66666666 when fees is 0 USDT
        assert order.created_last_price == decimal.Decimal(1500)
        assert order.is_filled()
        assert get_trade_fee_mock.call_count == 5
        trading_mode.create_order.assert_called_once()
        trading_mode.create_order.reset_mock()

    # using limit orders
    def _is_market_open_for_order_type(symbol: str, order_type: trading_enums.TraderOrderType):
        return (
            True if order_type in (trading_enums.TraderOrderType.SELL_LIMIT, trading_enums.TraderOrderType.BUY_LIMIT)
            else False
        )

    with mock.patch.object(
        exchange_manager.exchange, "is_market_open_for_order_type",
        mock.Mock(side_effect=_is_market_open_for_order_type)
    ) as is_market_open_for_order_type_mock:
        # cast 1: buying
        portfolio["USDT"] = trading_personal_data.SpotAsset("USDT", decimal.Decimal(1000), decimal.Decimal(1000))
        orders = await modes_util.convert_asset_to_target_asset(
            trading_mode, "USDT", "ETH", tickers, asset_amount=decimal.Decimal(450)
        )
        assert len(orders) == 1
        order = orders[0]
        assert order.order_type == trading_enums.TraderOrderType.BUY_LIMIT
        assert order.symbol == "ETH/USDT"
        # a bit lower than 0.3 because of the price change
        assert order.origin_quantity == decimal.Decimal("0.29850746")
        adapted_price = decimal.Decimal(1500) * (
            constants.ONE + constants.INSTANT_FILLED_LIMIT_ORDER_PRICE_DELTA
        )
        assert order.origin_price == adapted_price
        assert order.created_last_price == adapted_price
        # limit order is designed to be instantly filled, on simulator it should be filled
        assert order.is_filled()
        assert order.filled_price == order.origin_price
        assert order.filled_quantity > constants.ZERO
        assert is_market_open_for_order_type_mock.call_count == 2
        assert is_market_open_for_order_type_mock.mock_calls[0].args == \
               ("ETH/USDT", trading_enums.TraderOrderType.BUY_MARKET)
        assert is_market_open_for_order_type_mock.mock_calls[1].args == \
               ("ETH/USDT", trading_enums.TraderOrderType.BUY_LIMIT)
        is_market_open_for_order_type_mock.reset_mock()
        trading_mode.create_order.assert_called_once()
        trading_mode.create_order.reset_mock()

        # cast 2: selling
        portfolio["USDT"] = trading_personal_data.SpotAsset("USDT", decimal.Decimal(2), decimal.Decimal(2))
        orders = await modes_util.convert_asset_to_target_asset(
            trading_mode, "ETH", "USDT", tickers, asset_amount=decimal.Decimal(2)
        )
        assert len(orders) == 1
        order = orders[0]
        assert order.order_type == trading_enums.TraderOrderType.SELL_LIMIT
        assert order.symbol == "ETH/USDT"
        assert order.origin_quantity == decimal.Decimal("2")
        adapted_price = decimal.Decimal(1500) * (
            constants.ONE - constants.INSTANT_FILLED_LIMIT_ORDER_PRICE_DELTA
        )
        assert order.origin_price == adapted_price
        assert order.created_last_price == adapted_price
        assert order.is_filled()
        assert order.filled_price == order.origin_price
        assert order.filled_quantity > constants.ZERO
        assert is_market_open_for_order_type_mock.call_count == 2
        assert is_market_open_for_order_type_mock.mock_calls[0].args == \
               ("ETH/USDT", trading_enums.TraderOrderType.SELL_MARKET)
        assert is_market_open_for_order_type_mock.mock_calls[1].args == \
               ("ETH/USDT", trading_enums.TraderOrderType.SELL_LIMIT)
        is_market_open_for_order_type_mock.reset_mock()
        trading_mode.create_order.assert_called_once()
        trading_mode.create_order.reset_mock()


def _get_trading_mode(exchange_manager):
    async def _create_order(order, dependencies=None):
        return await signals.create_order(
            exchange_manager, False, order, dependencies=dependencies
        )

    return mock.Mock(
        exchange_manager=exchange_manager,
        create_order=mock.AsyncMock(side_effect=_create_order)
    )
